# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This module implements the base CCDPROC functions"""

import logging
import math
import numbers
import warnings

import numpy as np
from astropy import nddata, stats
from astropy import units as u
from astropy.modeling import fitting
from astropy.nddata import CCDData, StdDevUncertainty
from astropy.nddata.bitmask import (
    bitfield_to_boolean_mask as _bitfield_to_boolean_mask,
)
from astropy.units.quantity import Quantity
from astropy.utils import deprecated, deprecated_renamed_argument
from astropy.wcs.utils import proj_plane_pixel_area
from packaging import version as pkgversion
from scipy import ndimage

from .log_meta import log_to_metadata
from .utils.slices import slice_from_string

logger = logging.getLogger(__name__)

__all__ = [
    "background_deviation_box",
    "background_deviation_filter",
    "ccd_process",
    "cosmicray_median",
    "cosmicray_lacosmic",
    "create_deviation",
    "flat_correct",
    "gain_correct",
    "rebin",
    "sigma_func",
    "subtract_bias",
    "subtract_dark",
    "subtract_overscan",
    "transform_image",
    "trim_image",
    "wcs_project",
    "Keyword",
    "median_filter",
    "ccdmask",
    "bitfield_to_boolean_mask",
]

# The dictionary below is used to translate actual function names to names
# that are FITS compliant, i.e. 8 characters or less.
_short_names = {
    "background_deviation_box": "bakdevbx",
    "background_deviation_filter": "bakdfilt",
    "ccd_process": "ccdproc",
    "cosmicray_median": "crmedian",
    "create_deviation": "creatvar",
    "flat_correct": "flatcor",
    "gain_correct": "gaincor",
    "subtract_bias": "subbias",
    "subtract_dark": "subdark",
    "subtract_overscan": "suboscan",
    "trim_image": "trimim",
    "transform_image": "tranim",
    "wcs_project": "wcsproj",
}


@log_to_metadata
def ccd_process(
    ccd,
    oscan=None,
    trim=None,
    error=False,
    master_bias=None,
    dark_frame=None,
    master_flat=None,
    bad_pixel_mask=None,
    gain=None,
    readnoise=None,
    oscan_median=True,
    oscan_model=None,
    min_value=None,
    dark_exposure=None,
    data_exposure=None,
    exposure_key=None,
    exposure_unit=None,
    dark_scale=False,
    gain_corrected=True,
):
    """Perform basic processing on ccd data.

    The following steps can be included:

    * overscan correction (:func:`subtract_overscan`)
    * trimming of the image (:func:`trim_image`)
    * create deviation frame (:func:`create_deviation`)
    * gain correction (:func:`gain_correct`)
    * add a mask to the data
    * subtraction of master bias (:func:`subtract_bias`)
    * subtraction of a dark frame (:func:`subtract_dark`)
    * correction of flat field (:func:`flat_correct`)

    The task returns a processed `~astropy.nddata.CCDData` object.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Frame to be reduced.

    oscan : `~astropy.nddata.CCDData`, str or None, optional
        For no overscan correction, set to None. Otherwise provide a region
        of ccd from which the overscan is extracted, using the FITS
        conventions for index order and index start, or a
        slice from ccd that contains the overscan.
        Default is ``None``.

    trim : str or None, optional
        For no trim correction, set to None. Otherwise provide a region
        of ccd from which the image should be trimmed, using the FITS
        conventions for index order and index start.
        Default is ``None``.

    error : bool, optional
        If True, create an uncertainty array for ccd.
        Default is ``False``.

    master_bias : `~astropy.nddata.CCDData` or None, optional
        A master bias frame to be subtracted from ccd. The unit of the
        master bias frame should match the unit of the image **after
        gain correction** if ``gain_corrected`` is True.
        Default is ``None``.

    dark_frame : `~astropy.nddata.CCDData` or None, optional
        A dark frame to be subtracted from the ccd. The unit of the
        master dark frame should match the unit of the image **after
        gain correction** if ``gain_corrected`` is True.
        Default is ``None``.

    master_flat : `~astropy.nddata.CCDData` or None, optional
        A master flat frame to be divided into ccd. The unit of the
        master flat frame should match the unit of the image **after
        gain correction** if ``gain_corrected`` is True.
        Default is ``None``.

    bad_pixel_mask : `numpy.ndarray` or None, optional
        A bad pixel mask for the data. The bad pixel mask should be in given
        such that bad pixels have a value of 1 and good pixels a value of 0.
        Default is ``None``.

    gain : `~astropy.units.Quantity` or None, optional
        Gain value to multiple the image by to convert to electrons.
        Default is ``None``.

    readnoise : `~astropy.units.Quantity` or None, optional
        Read noise for the observations. The read noise should be in
        electrons.
        Default is ``None``.

    oscan_median : bool, optional
        If true, takes the median of each line. Otherwise, uses the mean.
        Default is ``True``.

    oscan_model : `~astropy.modeling.Model` or None, optional
        Model to fit to the data. If None, returns the values calculated
        by the median or the mean.
        Default is ``None``.

    min_value : float or None, optional
        Minimum value for flat field. The value can either be None and no
        minimum value is applied to the flat or specified by a float which
        will replace all values in the flat by the min_value.
        Default is ``None``.

    dark_exposure : `~astropy.units.Quantity` or None, optional
        Exposure time of the dark image; if specified, must also provided
        ``data_exposure``.
        Default is ``None``.

    data_exposure : `~astropy.units.Quantity` or None, optional
        Exposure time of the science image; if specified, must also provided
        ``dark_exposure``.
        Default is ``None``.

    exposure_key : `~ccdproc.Keyword`, str or None, optional
        Name of key in image metadata that contains exposure time.
        Default is ``None``.

    exposure_unit : `~astropy.units.Unit` or None, optional
        Unit of the exposure time if the value in the meta data does not
        include a unit.
        Default is ``None``.

    dark_scale : bool, optional
        If True, scale the dark frame by the exposure times.
        Default is ``False``.

    gain_corrected : bool, optional
        If True, the ``master_bias``, ``master_flat``, and ``dark_frame``
        have already been gain corrected.  Default is ``True``.

    Returns
    -------
    occd : `~astropy.nddata.CCDData`
        Reduded ccd.

    Examples
    --------
    1. To overscan, trim and gain correct a data set::

        >>> import numpy as np
        >>> from astropy import units as u
        >>> from astropy.nddata import CCDData
        >>> from ccdproc import ccd_process
        >>> ccd = CCDData(np.ones([100, 100]), unit=u.adu)
        >>> nccd = ccd_process(ccd, oscan='[1:10,1:100]',
        ...                    trim='[10:100, 1:100]', error=False,
        ...                    gain=2.0*u.electron/u.adu)
    """
    # make a copy of the object
    nccd = ccd.copy()

    # apply the overscan correction
    if isinstance(oscan, CCDData):
        nccd = subtract_overscan(
            nccd, overscan=oscan, median=oscan_median, model=oscan_model
        )
    elif isinstance(oscan, str):
        nccd = subtract_overscan(
            nccd, fits_section=oscan, median=oscan_median, model=oscan_model
        )
    elif oscan is None:
        pass
    else:
        raise TypeError("oscan is not None, a string, or CCDData object.")

    # apply the trim correction
    if isinstance(trim, str):
        nccd = trim_image(nccd, fits_section=trim)
    elif trim is None:
        pass
    else:
        raise TypeError("trim is not None or a string.")

    # create the error frame
    if error and gain is not None and readnoise is not None:
        nccd = create_deviation(nccd, gain=gain, readnoise=readnoise)
    elif error and (gain is None or readnoise is None):
        raise ValueError("gain and readnoise must be specified to create error frame.")

    # apply the bad pixel mask
    if isinstance(bad_pixel_mask, np.ndarray):
        nccd.mask = bad_pixel_mask
    elif bad_pixel_mask is None:
        pass
    else:
        raise TypeError("bad_pixel_mask is not None or numpy.ndarray.")

    # apply the gain correction
    if not (gain is None or isinstance(gain, Quantity)):
        raise TypeError("gain is not None or astropy.units.Quantity.")

    if gain is not None and gain_corrected:
        nccd = gain_correct(nccd, gain)

    # subtracting the master bias
    if isinstance(master_bias, CCDData):
        nccd = subtract_bias(nccd, master_bias)
    elif master_bias is None:
        pass
    else:
        raise TypeError("master_bias is not None or a CCDData object.")

    # subtract the dark frame
    if isinstance(dark_frame, CCDData):
        nccd = subtract_dark(
            nccd,
            dark_frame,
            dark_exposure=dark_exposure,
            data_exposure=data_exposure,
            exposure_time=exposure_key,
            exposure_unit=exposure_unit,
            scale=dark_scale,
        )
    elif dark_frame is None:
        pass
    else:
        raise TypeError("dark_frame is not None or a CCDData object.")

    # test dividing the master flat
    if isinstance(master_flat, CCDData):
        nccd = flat_correct(nccd, master_flat, min_value=min_value)
    elif master_flat is None:
        pass
    else:
        raise TypeError("master_flat is not None or a CCDData object.")

    # apply the gain correction only at the end if gain_corrected is False
    if gain is not None and not gain_corrected:
        nccd = gain_correct(nccd, gain)

    return nccd


@log_to_metadata
def create_deviation(ccd_data, gain=None, readnoise=None, disregard_nan=False):
    """
    Create a uncertainty frame. The function will update the uncertainty
    plane which gives the standard deviation for the data. Gain is used in
    this function only to scale the data in constructing the deviation; the
    data is not scaled.

    Parameters
    ----------
    ccd_data : `~astropy.nddata.CCDData`
        Data whose deviation will be calculated.

    gain : `~astropy.units.Quantity` or None, optional
        Gain of the CCD; necessary only if ``ccd_data`` and ``readnoise``
        are not in the same units. In that case, the units of ``gain``
        should be those that convert ``ccd_data.data`` to the same units as
        ``readnoise``.
        Default is ``None``.

    readnoise : `~astropy.units.Quantity` or None, optional
        Read noise per pixel.
        Default is ``None``.

    disregard_nan: boolean
        If ``True``, any value of nan in the output array will be replaced by
        the readnoise.

    {log}

    Raises
    ------
    UnitsError
        Raised if ``readnoise`` units are not equal to product of ``gain`` and
        ``ccd_data`` units.

    Returns
    -------
    ccd : `~astropy.nddata.CCDData`
        CCDData object with uncertainty created; uncertainty is in the same
        units as the data in the parameter ``ccd_data``.

    """
    if gain is not None and not isinstance(gain, Quantity):
        raise TypeError("gain must be a astropy.units.Quantity.")

    if readnoise is None:
        raise ValueError("must provide a readnoise.")

    if not isinstance(readnoise, Quantity):
        raise TypeError("readnoise must be a astropy.units.Quantity.")

    if gain is None:
        gain = 1.0 * u.dimensionless_unscaled

    if gain.unit * ccd_data.unit != readnoise.unit:
        raise u.UnitsError("units of data, gain and readnoise do not match.")

    # Need to convert Quantity to plain number because NDData data is not
    # a Quantity. All unit checking should happen prior to this point.
    gain_value = float(gain / gain.unit)
    readnoise_value = float(readnoise / readnoise.unit)

    # remove values that might be negative or treat as nan
    data = gain_value * ccd_data.data
    mask = data < 0
    if disregard_nan:
        data[mask] = 0
    else:
        data[mask] = np.nan
        logging.warning("Negative values in array will be replaced with nan")

    # calculate the deviation
    var = (data + readnoise_value**2) ** 0.5

    # ensure uncertainty and image data have same unit
    ccd = ccd_data.copy()
    var /= gain_value
    ccd.uncertainty = StdDevUncertainty(var)
    return ccd


@log_to_metadata
def subtract_overscan(
    ccd, overscan=None, overscan_axis=1, fits_section=None, median=False, model=None
):
    """
    Subtract the overscan region from an image.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Data to have overscan frame corrected.

    overscan : `~astropy.nddata.CCDData` or None, optional
        Slice from ``ccd`` that contains the overscan. Must provide either
        this argument or ``fits_section``, but not both.
        Default is ``None``.

    overscan_axis : 0, 1 or None, optional
        Axis along which overscan should combined with mean or median. Axis
        numbering follows the *python* convention for ordering, so 0 is the
        first axis and 1 is the second axis.

        If overscan_axis is explicitly set to None, the axis is set to
        the shortest dimension of the overscan section (or 1 in case
        of a square overscan).
        Default is ``1``.

    fits_section : str or None, optional
        Region of ``ccd`` from which the overscan is extracted, using the FITS
        conventions for index order and index start. See Notes and Examples
        below. Must provide either this argument or ``overscan``, but not both.
        Default is ``None``.

    median : bool, optional
        If true, takes the median of each line. Otherwise, uses the mean.
        Default is ``False``.

    model : `~astropy.modeling.Model` or None, optional
        Model to fit to the data. If None, returns the values calculated
        by the median or the mean.
        Default is ``None``.

    {log}

    Raises
    ------
    TypeError
        A TypeError is raised if either ``ccd`` or ``overscan`` are not the
        correct objects.

    Returns
    -------
    ccd : `~astropy.nddata.CCDData`
        CCDData object with overscan subtracted.

    Notes
    -----
    The format of the ``fits_section`` string follow the rules for slices that
    are consistent with the FITS standard (v3) and IRAF usage of keywords like
    TRIMSEC and BIASSEC. Its indexes are one-based, instead of the
    python-standard zero-based, and the first index is the one that increases
    most rapidly as you move through the array in memory order, opposite the
    python ordering.

    The 'fits_section' argument is provided as a convenience for those who are
    processing files that contain TRIMSEC and BIASSEC. The preferred, more
    pythonic, way of specifying the overscan is to do it by indexing the data
    array directly with the ``overscan`` argument.

    Examples
    --------
    Creating a 100x100 array containing ones just for demonstration purposes::

        >>> import numpy as np
        >>> from astropy import units as u
        >>> arr1 = CCDData(np.ones([100, 100]), unit=u.adu)

    The statement below uses all rows of columns 90 through 99 as the
    overscan::

        >>> no_scan = subtract_overscan(arr1, overscan=arr1[:, 90:100])
        >>> assert (no_scan.data == 0).all()

    This statement does the same as the above, but with a FITS-style section::

        >>> no_scan = subtract_overscan(arr1, fits_section='[91:100, :]')
        >>> assert (no_scan.data == 0).all()

    Spaces are stripped out of the ``fits_section`` string.

    """
    if not isinstance(ccd, CCDData):
        raise TypeError("ccddata is not a CCDData object.")

    if (overscan is not None and fits_section is not None) or (
        overscan is None and fits_section is None
    ):
        raise TypeError("specify either overscan or fits_section, but not " "both.")

    if (overscan is not None) and (not isinstance(overscan, CCDData)):
        raise TypeError("overscan is not a CCDData object.")

    if fits_section is not None and not isinstance(fits_section, str):
        raise TypeError("overscan is not a string.")

    if fits_section is not None:
        overscan = ccd[slice_from_string(fits_section, fits_convention=True)]

    if overscan_axis is None:
        overscan_axis = 0 if overscan.shape[1] > overscan.shape[0] else 1

    if median:
        oscan = np.median(overscan.data, axis=overscan_axis)
    else:
        oscan = np.mean(overscan.data, axis=overscan_axis)

    if model is not None:
        of = fitting.LinearLSQFitter()
        yarr = np.arange(len(oscan))
        oscan = of(model, yarr, oscan)
        oscan = oscan(yarr)
        if overscan_axis == 1:
            oscan = np.reshape(oscan, (oscan.size, 1))
        else:
            oscan = np.reshape(oscan, (1, oscan.size))
    else:
        if overscan_axis == 1:
            oscan = np.reshape(oscan, oscan.shape + (1,))
        else:
            oscan = np.reshape(oscan, (1,) + oscan.shape)

    subtracted = ccd.copy()

    # subtract the overscan
    subtracted.data = ccd.data - oscan
    return subtracted


@log_to_metadata
def trim_image(ccd, fits_section=None):
    """
    Trim the image to the dimensions indicated.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        CCD image to be trimmed, sliced if desired.

    fits_section : str or None, optional
        Region of ``ccd`` from which the overscan is extracted; see
        `~ccdproc.subtract_overscan` for details.
        Default is ``None``.

    {log}

    Returns
    -------
    trimmed_ccd : `~astropy.nddata.CCDData`
        Trimmed image.

    Examples
    --------
    Given an array that is 100x100,

        >>> import numpy as np
        >>> from astropy import units as u
        >>> arr1 = CCDData(np.ones([100, 100]), unit=u.adu)

    the syntax for trimming this to keep all of the first index but only the
    first 90 rows of the second index is

        >>> trimmed = trim_image(arr1[:, :90])
        >>> trimmed.shape
        (100, 90)
        >>> trimmed.data[0, 0] = 2
        >>> float(arr1.data[0, 0])
        1.0

    This both trims *and makes a copy* of the image.

    Indexing the image directly does *not* do the same thing, quite:

        >>> not_really_trimmed = arr1[:, :90]
        >>> not_really_trimmed.data[0, 0] = 2
        >>> float(arr1.data[0, 0])
        2.0

    In this case, ``not_really_trimmed`` is a view of the underlying array
    ``arr1``, not a copy.
    """
    if fits_section is not None and not isinstance(fits_section, str):
        raise TypeError("fits_section must be a string.")
    trimmed = ccd.copy()
    if fits_section:
        python_slice = slice_from_string(fits_section, fits_convention=True)
        trimmed = trimmed[python_slice]

    return trimmed


@log_to_metadata
def subtract_bias(ccd, master):
    """
    Subtract master bias from image.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Image from which bias will be subtracted.

    master : `~astropy.nddata.CCDData`
        Master image to be subtracted from ``ccd``.

    {log}

    Returns
    -------
    result : `~astropy.nddata.CCDData`
        CCDData object with bias subtracted.
    """

    try:
        result = ccd.subtract(master)
    except ValueError as err:
        if "operand units" in str(err):
            raise u.UnitsError(
                f"Unit '{ccd.unit}' of the uncalibrated image does not "
                f"match unit '{master.unit}' of the calibration "
                "image"
            ) from err
        else:
            raise err

    result.meta = ccd.meta.copy()
    return result


@log_to_metadata
def subtract_dark(
    ccd,
    master,
    dark_exposure=None,
    data_exposure=None,
    exposure_time=None,
    exposure_unit=None,
    scale=False,
):
    """
    Subtract dark current from an image.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Image from which dark will be subtracted.

    master : `~astropy.nddata.CCDData`
        Dark image.

    dark_exposure : `~astropy.units.Quantity` or None, optional
        Exposure time of the dark image; if specified, must also provided
        ``data_exposure``.
        Default is ``None``.

    data_exposure : `~astropy.units.Quantity` or None, optional
        Exposure time of the science image; if specified, must also provided
        ``dark_exposure``.
        Default is ``None``.

    exposure_time : str or `~ccdproc.Keyword` or None, optional
        Name of key in image metadata that contains exposure time.
        Default is ``None``.

    exposure_unit : `~astropy.units.Unit` or None, optional
        Unit of the exposure time if the value in the meta data does not
        include a unit.
        Default is ``None``.

    scale: bool, optional
        If True, scale the dark frame by the exposure times.
        Default is ``False``.

    {log}

    Returns
    -------
    result : `~astropy.nddata.CCDData`
        Dark-subtracted image.
    """
    if ccd.shape != master.shape:
        err_str = (
            f"operands could not be subtracted with "
            f"shapes {ccd.shape} {master.shape}"
        )
        raise ValueError(err_str)

    if not (isinstance(ccd, CCDData) and isinstance(master, CCDData)):
        raise TypeError("ccd and master must both be CCDData objects.")

    if (
        data_exposure is not None
        and dark_exposure is not None
        and exposure_time is not None
    ):
        raise TypeError(
            "specify either exposure_time or "
            "(dark_exposure and data_exposure), not both."
        )

    if data_exposure is None and dark_exposure is None:
        if exposure_time is None:
            raise TypeError(
                "must specify either exposure_time or both "
                "dark_exposure and data_exposure."
            )
        if isinstance(exposure_time, Keyword):
            data_exposure = exposure_time.value_from(ccd.header)
            dark_exposure = exposure_time.value_from(master.header)
        else:
            data_exposure = ccd.header[exposure_time]
            dark_exposure = master.header[exposure_time]

    if not (
        isinstance(dark_exposure, Quantity) and isinstance(data_exposure, Quantity)
    ):
        if exposure_time:
            try:
                data_exposure *= exposure_unit
                dark_exposure *= exposure_unit
            except TypeError as err:
                raise TypeError("must provide unit for exposure time.") from err
        else:
            raise TypeError("exposure times must be astropy.units.Quantity objects.")

    try:
        if scale:
            master_scaled = master.copy()
            # data_exposure and dark_exposure are both quantities,
            # so we can just have subtract do the scaling
            master_scaled = master_scaled.multiply(data_exposure / dark_exposure)
            result = ccd.subtract(master_scaled)
        else:
            result = ccd.subtract(master)
    except (u.UnitsError, u.UnitConversionError, ValueError) as err:
        # Make the error message a little more explicit than what is returned
        # by default.
        raise u.UnitsError(
            f"Unit '{ccd.unit}' of the uncalibrated image does not "
            f"match unit '{master.unit}' of the calibration "
            "image"
        ) from err

    result.meta = ccd.meta.copy()
    return result


@log_to_metadata
def gain_correct(ccd, gain, gain_unit=None):
    """Correct the gain in the image.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
      Data to have gain corrected.

    gain : `~astropy.units.Quantity` or `~ccdproc.Keyword`
      gain value for the image expressed in electrons per adu.

    gain_unit : `~astropy.units.Unit` or None, optional
        Unit for the ``gain``; used only if ``gain`` itself does not provide
        units.
        Default is ``None``.

    {log}

    Returns
    -------
    result : `~astropy.nddata.CCDData`
      CCDData object with gain corrected.
    """
    if isinstance(gain, Keyword):
        gain_value = gain.value_from(ccd.header)
    elif isinstance(gain, numbers.Number) and gain_unit is not None:
        gain_value = gain * u.Unit(gain_unit)
    else:
        gain_value = gain

    result = ccd.multiply(gain_value)
    result.meta = ccd.meta.copy()
    return result


@log_to_metadata
def flat_correct(ccd, flat, min_value=None, norm_value=None):
    """Correct the image for flat fielding.

    The flat field image is normalized by its mean or a user-supplied value
    before flat correcting.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Data to be transformed.

    flat : `~astropy.nddata.CCDData`
        Flatfield to apply to the data.

    min_value : float or None, optional
        Minimum value for flat field. The value can either be None and no
        minimum value is applied to the flat or specified by a float which
        will replace all values in the flat by the min_value.
        Default is ``None``.

    norm_value : float or None, optional
        If not ``None``, normalize flat field by this argument rather than the
        mean of the image. This allows fixing several different flat fields to
        have the same scale. If this value is negative or 0, a ``ValueError``
        is raised. Default is ``None``.

    {log}

    Returns
    -------
    ccd : `~astropy.nddata.CCDData`
        CCDData object with flat corrected.
    """
    # Use the min_value to replace any values in the flat
    use_flat = flat
    if min_value is not None:
        flat_min = flat.copy()
        flat_min.data[flat_min.data < min_value] = min_value
        use_flat = flat_min

    # If a norm_value was input and is positive, use it to scale the flat
    if norm_value is not None and norm_value > 0:
        flat_mean = (
            norm_value if hasattr(norm_value, "unit") else norm_value * use_flat.unit
        )
    elif norm_value is not None:
        # norm_value was set to a bad value
        raise ValueError("norm_value must be greater than zero.")
    else:
        # norm_value was not set, use mean of the image.
        flat_mean = use_flat.data.mean() * use_flat.unit

    # Normalize the flat.
    flat_normed = use_flat.divide(flat_mean)

    # Set masked values to unity; the array element remains masked, but the data
    # value is set to unity to avoid runtime divide-by-zero errors that are due
    # to a masked value being set to 0.
    if flat_normed.mask is not None and flat_normed.mask.any():
        flat_normed.data[flat_normed.mask] = 1.0

    # divide through the flat
    flat_corrected = ccd.divide(flat_normed)

    flat_corrected.meta = ccd.meta.copy()
    return flat_corrected


@log_to_metadata
def transform_image(ccd, transform_func, **kwargs):
    """Transform the image.

    Using the function specified by transform_func, the transform will
    be applied to data, uncertainty, and mask in ccd.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Data to be transformed.

    transform_func : callable
        Function to be used to transform the data, mask and uncertainty.

    kwargs :
        Additional keyword arguments to be used by the transform_func.

    {log}

    Returns
    -------
    ccd : `~astropy.nddata.CCDData`
        A transformed CCDData object.

    Notes
    -----
    At this time, transform will be applied to the uncertainty data but it
    will only transform the data. This will not properly handle uncertainties
    that arise due to correlation between the pixels.

    These should only be geometric transformations of the images. Other
    methods should be used if the units of ccd need to be changed.

    Examples
    --------
    Given an array that is 100x100::

        >>> import numpy as np
        >>> from astropy import units as u
        >>> arr1 = CCDData(np.ones([100, 100]), unit=u.adu)

    The syntax for transforming the array using
    `scipy.ndimage.shift`::

        >>> from scipy.ndimage import shift
        >>> from ccdproc import transform_image
        >>> transformed = transform_image(arr1, shift, shift=(5.5, 8.1))
    """
    # check that it is a ccddata object
    if not isinstance(ccd, CCDData):
        raise TypeError("ccd is not a CCDData.")

    # make a copy of the object
    nccd = ccd.copy()

    # transform the image plane
    try:
        nccd.data = transform_func(nccd.data, **kwargs)
    except TypeError as exc:
        if "is not callable" in str(exc):
            raise TypeError("transform_func is not a callable.") from exc
        raise

    # transform the uncertainty plane if it exists
    if nccd.uncertainty is not None:
        nccd.uncertainty.array = transform_func(nccd.uncertainty.array, **kwargs)

    # transform the mask plane
    if nccd.mask is not None:
        mask = transform_func(nccd.mask, **kwargs)
        nccd.mask = mask > 0

    if nccd.wcs is not None:
        warn = "WCS information may be incorrect as no transformation was applied to it"
        warnings.warn(warn, UserWarning, stacklevel=2)

    return nccd


@log_to_metadata
def wcs_project(ccd, target_wcs, target_shape=None, order="bilinear"):
    """
    Given a CCDData image with WCS, project it onto a target WCS and
    return the reprojected data as a new CCDData image.

    Any flags, weight, or uncertainty are ignored in doing the
    reprojection.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`
        Data to be projected.

    target_wcs : `~astropy.wcs.WCS` object
        WCS onto which all images should be projected.

    target_shape : two element list-like or None, optional
        Shape of the output image. If omitted, defaults to the shape of the
        input image.
        Default is ``None``.

    order : str, optional
        Interpolation order for re-projection. Must be one of:

        + 'nearest-neighbor'
        + 'bilinear'
        + 'biquadratic'
        + 'bicubic'

        Default is ``'bilinear'``.

    {log}

    Returns
    -------
    ccd : `~astropy.nddata.CCDData`
        A transformed CCDData object.
    """
    from astropy.nddata.ccddata import _generate_wcs_and_update_header
    from reproject import reproject_interp

    if not (ccd.wcs.is_celestial and target_wcs.is_celestial):
        raise ValueError("one or both WCS is not celestial.")

    if target_shape is None:
        target_shape = ccd.shape

    projected_image_raw, _ = reproject_interp(
        (ccd.data, ccd.wcs), target_wcs, shape_out=target_shape, order=order
    )

    reprojected_mask = None
    if ccd.mask is not None:
        reprojected_mask, _ = reproject_interp(
            (ccd.mask, ccd.wcs), target_wcs, shape_out=target_shape, order=order
        )
        # Make the mask 1 if the reprojected mask pixel value is non-zero.
        # A small threshold is included to allow for some rounding in
        # reproject_interp.
        reprojected_mask = reprojected_mask > 1e-8

    # The reprojection will contain nan for any pixels for which the source
    # was outside the original image. Those should be masked also.
    output_mask = np.isnan(projected_image_raw)

    if reprojected_mask is not None:
        output_mask = output_mask | reprojected_mask

    # Need to scale counts by ratio of pixel areas
    area_ratio = proj_plane_pixel_area(target_wcs) / proj_plane_pixel_area(ccd.wcs)

    # If nothing ended up masked, don't create a mask.
    if not output_mask.any():
        output_mask = None

    # If there are any wcs keywords in the header, remove them
    hdr, _ = _generate_wcs_and_update_header(ccd.header)

    nccd = CCDData(
        area_ratio * projected_image_raw,
        wcs=target_wcs,
        mask=output_mask,
        header=hdr,
        unit=ccd.unit,
    )

    return nccd


def sigma_func(arr, axis=None, ignore_nan=False):
    """
    Robust method for calculating the deviation of an array. ``sigma_func``
    uses the median absolute deviation to determine the standard deviation.

    Parameters
    ----------
    arr : `~astropy.nddata.CCDData` or `numpy.ndarray`
        Array whose deviation is to be calculated.

    axis : int, tuple of ints or None, optional
        Axis or axes along which the function is performed.
        If ``None`` it is performed over all the dimensions of
        the input array. The axis argument can also be negative, in this case
        it counts from the last to the first axis.
        Default is ``None``.

    Returns
    -------
    uncertainty : float
        uncertainty of array estimated from median absolute deviation.
    """
    return (
        stats.median_absolute_deviation(arr, axis=axis, ignore_nan=ignore_nan)
        * 1.482602218505602
    )


def setbox(x, y, mbox, xmax, ymax):
    """
    Create a box of length mbox around a position x,y. If the box will
    be out of [0,len] then reset the edges of the box to be within the
    boundaries.

    Parameters
    ----------
    x : int
        Central x-position of box.

    y : int
        Central y-position of box.

    mbox : int
        Width of box.

    xmax : int
        Maximum x value.

    ymax : int
        Maximum y value.

    Returns
    -------
    x1 : int
        Lower x corner of box.

    x2 : int
        Upper x corner of box.

    y1 : int
        Lower y corner of box.

    y2 : int
        Upper y corner of box.
    """
    mbox = max(int(0.5 * mbox), 1)
    y1 = max(0, y - mbox)
    y2 = min(y + mbox + 1, ymax - 1)
    x1 = max(0, x - mbox)
    x2 = min(x + mbox + 1, xmax - 1)

    return x1, x2, y1, y2


def background_deviation_box(data, bbox):
    """
    Determine the background deviation with a box size of bbox. The algorithm
    steps through the image and calculates the deviation within each box.
    It returns an array with the pixels in each box filled with the deviation
    value.

    Parameters
    ----------
    data : `numpy.ndarray` or `numpy.ma.MaskedArray`
        Data to measure background deviation.

    bbox : int
        Box size for calculating background deviation.

    Raises
    ------
    ValueError
        A value error is raised if bbox is less than 1.

    Returns
    -------
    background : `numpy.ndarray` or `numpy.ma.MaskedArray`
        An array with the measured background deviation in each pixel.
    """
    # Check to make sure the background box is an appropriate size
    # If it is too small, then insufficient statistics are generated
    if bbox < 1:
        raise ValueError("bbox must be greater than 1.")

    # make the background image
    barr = data * 0.0 + data.std()
    ylen, xlen = data.shape
    for i in range(int(0.5 * bbox), xlen, bbox):
        for j in range(int(0.5 * bbox), ylen, bbox):
            x1, x2, y1, y2 = setbox(i, j, bbox, xlen, ylen)
            barr[y1:y2, x1:x2] = sigma_func(data[y1:y2, x1:x2])

    return barr


def background_deviation_filter(data, bbox):
    """
    Determine the background deviation for each pixel from a box with size of
    bbox.

    Parameters
    ----------
    data : `numpy.ndarray`
        Data to measure background deviation.

    bbox : int
        Box size for calculating background deviation.

    Raises
    ------
    ValueError
        A value error is raised if bbox is less than 1.

    Returns
    -------
    background : `numpy.ndarray` or `numpy.ma.MaskedArray`
        An array with the measured background deviation in each pixel.
    """
    # Check to make sure the background box is an appropriate size
    if bbox < 1:
        raise ValueError("bbox must be greater than 1.")

    return ndimage.generic_filter(data, sigma_func, size=(bbox, bbox))


@deprecated(
    "1.1",
    message="The rebin function will be removed in ccdproc 3.0 "
    "Use block_reduce or block_replicate instead.",
)
def rebin(ccd, newshape):
    """
    Rebin an array to have a new shape.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData` or `numpy.ndarray`
        Data to rebin.

    newshape : tuple
        Tuple containing the new shape for the array.

    Returns
    -------
    output : `~astropy.nddata.CCDData` or `numpy.ndarray`
        An array with the new shape. It will have the same type as the input
        object.

    Raises
    ------
    TypeError
        A type error is raised if data is not an `numpy.ndarray` or
        `~astropy.nddata.CCDData`.

    ValueError
        A value error is raised if the dimension of the new shape is not equal
        to the data's.

    Notes
    -----
    This is based on the scipy cookbook for rebinning:
    http://wiki.scipy.org/Cookbook/Rebinning

    If rebinning a CCDData object to a smaller shape, the masking and
    uncertainty are not handled correctly.

    Examples
    --------
    Given an array that is 100x100::

        import numpy as np
        from astropy import units as u
        arr1 = CCDData(np.ones([10, 10]), unit=u.adu)

    The syntax for rebinning an array to a shape
    of (20,20) is::

        rebin(arr1, (20,20))
    """
    # check to see that is in a nddata type
    if isinstance(ccd, np.ndarray):

        # check to see that the two arrays are going to be the same length
        if len(ccd.shape) != len(newshape):
            raise ValueError("newshape does not have the same dimensions as " "ccd.")

        slices = [slice(0, old, old / new) for old, new in zip(ccd.shape, newshape)]
        coordinates = np.mgrid[slices]
        indices = coordinates.astype("i")
        return ccd[tuple(indices)]

    elif isinstance(ccd, CCDData):
        # check to see that the two arrays are going to be the same length
        if len(ccd.shape) != len(newshape):
            raise ValueError("newshape does not have the same dimensions as ccd.")

        nccd = ccd.copy()
        # rebin the data plane
        nccd.data = rebin(nccd.data, newshape)

        # rebin the uncertainty plane
        if nccd.uncertainty is not None:
            nccd.uncertainty.array = rebin(nccd.uncertainty.array, newshape)

        # rebin the mask plane
        if nccd.mask is not None:
            nccd.mask = rebin(nccd.mask, newshape)

        return nccd
    else:
        raise TypeError("ccd is not an ndarray or a CCDData object.")


def block_reduce(ccd, block_size, func=np.sum):
    """Thin wrapper around `astropy.nddata.block_reduce`."""
    data = nddata.block_reduce(ccd, block_size, func)
    if isinstance(ccd, CCDData):
        # unit and meta "should" be unaffected by the change of shape and can
        # be copied. However wcs, mask, uncertainty should not be copied!
        data = CCDData(data, unit=ccd.unit, meta=ccd.meta.copy())
    return data


def block_average(ccd, block_size):
    """Like `block_reduce` but with predefined ``func=np.mean``."""
    data = nddata.block_reduce(ccd, block_size, np.mean)
    # Like in block_reduce:
    if isinstance(ccd, CCDData):
        data = CCDData(data, unit=ccd.unit, meta=ccd.meta.copy())
    return data


def block_replicate(ccd, block_size, conserve_sum=True):
    """Thin wrapper around `astropy.nddata.block_replicate`."""
    data = nddata.block_replicate(ccd, block_size, conserve_sum)
    # Like in block_reduce:
    if isinstance(ccd, CCDData):
        data = CCDData(data, unit=ccd.unit, meta=ccd.meta.copy())
    return data


try:
    # Append original docstring to docstrings of these functions
    block_reduce.__doc__ += nddata.block_reduce.__doc__
    block_replicate.__doc__ += nddata.block_replicate.__doc__
    __all__ += ["block_average", "block_reduce", "block_replicate"]
except AttributeError:
    # Astropy 1.0 has no block_reduce, block_average
    del block_reduce, block_average, block_replicate


def median_filter(data, *args, **kwargs):
    """See `scipy.ndimage.median_filter` for arguments.

    If the ``data`` is a `~astropy.nddata.CCDData` object the result will be another
    `~astropy.nddata.CCDData` object with the median filtered data as ``data`` and
    copied ``unit`` and ``meta``.
    """
    if isinstance(data, CCDData):
        out_kwargs = {"meta": data.meta.copy(), "unit": data.unit}
        result = ndimage.median_filter(data.data, *args, **kwargs)
        return CCDData(result, **out_kwargs)
    else:
        return ndimage.median_filter(data, *args, **kwargs)


# This originally used the "message" argument but that is not
# supported until astropy 5, so use alternative instead.
@deprecated_renamed_argument(
    "pssl",
    None,
    "2.3.0",
    arg_in_kwargs=True,
    alternative="The pssl keyword will be removed in "
    "ccdproc 3.0. Use inbkg instead to have astroscrappy temporarily remove the "
    "background during processing.",
)
def cosmicray_lacosmic(
    ccd,
    sigclip=4.5,
    sigfrac=0.3,
    objlim=5.0,
    gain=1.0,
    readnoise=6.5,
    satlevel=65535.0,
    pssl=0.0,
    niter=4,
    sepmed=True,
    cleantype="meanmask",
    fsmode="median",
    psfmodel="gauss",
    psffwhm=2.5,
    psfsize=7,
    psfk=None,
    psfbeta=4.765,
    verbose=False,
    gain_apply=True,
    inbkg=None,
    invar=None,
):
    r"""
    Identify cosmic rays through the L.A. Cosmic technique. The L.A. Cosmic
    technique identifies cosmic rays by identifying pixels based on a variation
    of the Laplacian edge detection. The algorithm is an implementation of the
    code describe in van Dokkum (2001) [1]_ as implemented by McCully (2014)
    [2]_. If you use this algorithm, please cite these two works.


    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData` or `numpy.ndarray`
        Data to have cosmic ray cleaned.

    gain_apply : bool, optional
        If ``True``, **return gain-corrected data**, with correct units,
        otherwise do not gain-correct the data. Default is ``True`` to
        preserve backwards compatibility.

    sigclip : float, optional
        Laplacian-to-noise limit for cosmic ray detection. Lower values will
        flag more pixels as cosmic rays. Default: 4.5.

    sigfrac : float, optional
        Fractional detection limit for neighboring pixels. For cosmic ray
        neighbor pixels, a Laplacian-to-noise detection limit of
        sigfrac * sigclip will be used. Default: 0.3.

    objlim : float, optional
        Minimum contrast between Laplacian image and the fine structure image.
        Increase this value if cores of bright stars are flagged as cosmic
        rays. Default: 5.0.

    inbkg : float numpy array, optional
        A pre-determined background image, to be subtracted from ``indat``
        before running the main detection algorithm.
        This is used primarily with spectroscopic data, to remove
        sky lines and the cross-section of an object continuum during
        iteration, "protecting" them from spurious rejection (see the above
        paper). This background is not removed from the final, cleaned output
        (``cleanarr``). This should be in units of "counts", the same units of indat.
        This inbkg should be free from cosmic rays. When estimating the cosmic-ray
        free noise of the image, we will treat ``inbkg`` as a constant Poisson
        contribution to the variance.

    invar : float numpy array, optional
        A pre-determined estimate of the data variance (ie. noise squared) in
        each pixel, generated by previous processing of ``indat``. If provided,
        this is used in place of an internal noise model based on ``indat``,
        ``gain`` and ``readnoise``. This still gets median filtered and cleaned
        internally, to estimate what the noise in each pixel *would* be in the
        absence of cosmic rays. This should be in units of "counts" squared.

    pssl : float, optional
        Previously subtracted sky level in ADU. We always need to work in
        electrons for cosmic ray detection, so we need to know the sky level
        that has been subtracted so we can add it back in. Default: 0.0.

    gain : float or `~astropy.units.Quantity`, optional
        Gain of the image (electrons / ADU). We always need to work in
        electrons for cosmic ray detection. Default: 1.0

    readnoise : float, optional
        Read noise of the image (electrons). Used to generate the noise model
        of the image. Default: 6.5.

    satlevel : float, optional
        Saturation level of the image (electrons). This value is used to
        detect saturated stars and pixels at or above this level are added to
        the mask. Default: 65535.0.

    niter : int, optional
        Number of iterations of the LA Cosmic algorithm to perform. Default: 4.

    sepmed : bool, optional
        Use the separable median filter instead of the full median filter.
        The separable median is not identical to the full median filter, but
        they are approximately the same, the separable median filter is
        significantly faster, and still detects cosmic rays well. Note, this is
        a performance feature, and not part of the original L.A. Cosmic.
        Default: True

    cleantype : str, optional
        Set which clean algorithm is used:

        - ``"median"``: An unmasked 5x5 median filter.
        - ``"medmask"``: A masked 5x5 median filter.
        - ``"meanmask"``: A masked 5x5 mean filter.
        - ``"idw"``: A masked 5x5 inverse distance weighted interpolation.

        Default: ``"meanmask"``.

    fsmode : str, optional
        Method to build the fine structure image:

        - ``"median"``: Use the median filter in the standard LA Cosmic \
          algorithm.
        - ``"convolve"``: Convolve the image with the psf kernel to calculate \
          the fine structure image.

        Default: ``"median"``.

    psfmodel : str, optional
        Model to use to generate the psf kernel if fsmode == 'convolve' and
        psfk is None. The current choices are Gaussian and Moffat profiles:

        - ``"gauss"`` and ``"moffat"`` produce circular PSF kernels.
        - The ``"gaussx"`` and ``"gaussy"`` produce Gaussian kernels in the x \
          and y directions respectively.

        Default: ``"gauss"``.

    psffwhm : float, optional
        Full Width Half Maximum of the PSF to use to generate the kernel.
        Default: 2.5.

    psfsize : int, optional
        Size of the kernel to calculate. Returned kernel will have size
        psfsize x psfsize. psfsize should be odd. Default: 7.

    psfk : `numpy.ndarray` (with float dtype) or None, optional
        PSF kernel array to use for the fine structure image if
        ``fsmode == 'convolve'``. If None and ``fsmode == 'convolve'``, we
        calculate the psf kernel using ``psfmodel``. Default: None.

    psfbeta : float, optional
        Moffat beta parameter. Only used if ``fsmode=='convolve'`` and
        ``psfmodel=='moffat'``. Default: 4.765.

    verbose : bool, optional
        Print to the screen or not. Default: False.

    Notes
    -----
    Implementation of the cosmic ray identification L.A.Cosmic:
    http://www.astro.yale.edu/dokkum/lacosmic/

    Returns
    -------
    nccd : `~astropy.nddata.CCDData` or `numpy.ndarray`
        An object of the same type as ccd is returned. If it is a
        `~astropy.nddata.CCDData`, the mask attribute will also be updated with
        areas identified with cosmic rays masked. **By default, the image is
        multiplied by the gain.** You can control this behavior with the
        ``gain_apply`` argument.

    crmask : `numpy.ndarray`
        If an `numpy.ndarray` is provided as ccd, a boolean ndarray with the
        cosmic rays identified will also be returned.

    References
    ----------
    .. [1] van Dokkum, P; 2001, "Cosmic-Ray Rejection by Laplacian Edge
       Detection". The Publications of the Astronomical Society of the
       Pacific, Volume 113, Issue 789, pp. 1420-1427.
       doi: 10.1086/323894

    .. [2] McCully, C., 2014, "Astro-SCRAPPY",
       https://github.com/astropy/astroscrappy

    Examples
    --------
    1) Given an numpy.ndarray object, the syntax for running
       cosmicrar_lacosmic would be:

       >>> newdata, mask = cosmicray_lacosmic(data, sigclip=5)  #doctest: +SKIP

       where the error is an array that is the same shape as data but
       includes the pixel error. This would return a data array, newdata,
       with the bad pixels replaced by the local median from a box of 11
       pixels; and it would return a mask indicating the bad pixels.

    2) Given an `~astropy.nddata.CCDData` object with an uncertainty frame, the syntax
       for running cosmicrar_lacosmic would be:

       >>> newccd = cosmicray_lacosmic(ccd, sigclip=5)   # doctest: +SKIP

       The newccd object will have bad pixels in its data array replace and the
       mask of the object will be created if it did not previously exist or be
       updated with the detected cosmic rays.
    """
    from astroscrappy import __version__ as asy_version
    from astroscrappy import detect_cosmics

    # If we didn't get a quantity, put them in, with unit specified by the
    # documentation above.
    if not isinstance(gain, u.Quantity):
        # Gain will change the value, so use the proper units
        gain = gain * u.electron / u.adu

    # Set the units of readnoise to electrons, as specified in the
    # documentation, if no unit is present.
    if not isinstance(readnoise, u.Quantity):
        readnoise = readnoise * u.electron

    # Handle transition from old astroscrappy interface to new
    old_astroscrappy_interface = pkgversion.parse(asy_version) < pkgversion.parse(
        "1.1.0"
    )

    # Use this dictionary to define which keyword arguments are actually
    # passed to astroscrappy.
    asy_background_kwargs = {}

    # This is for handling the transition in astroscrappy versions
    data_offset = 0

    # Handle setting up the keyword arguments for both interfaces
    if old_astroscrappy_interface:  # pragma: no cover
        new_args = dict(inbkg=inbkg, invar=invar)
        bad_args = []
        for k, v in new_args.items():
            if v is not None:
                bad_args.append(k)

        if bad_args:
            s = "s" if len(bad_args) > 1 else ""
            bads = ", ".join(bad_args)
            raise TypeError(
                f"The argument{s} {bads} only valid for astroscrappy "
                "1.1.0 or higher."
            )

        if pssl != 0:
            asy_background_kwargs = dict(pssl=pssl)

    else:
        if pssl != 0:
            if (inbkg is not None) or (invar is not None):
                raise ValueError("Cannot set both pssl and inbkg")

            # The old version of astroscrappy added the bkg back in
            # if pssl was provided. The new one does not, so set an offset
            # here that we later add in then take out.
            data_offset = pssl

        asy_background_kwargs = dict(inbkg=inbkg, invar=invar)

    if isinstance(ccd, np.ndarray):
        data = ccd

        crmask, cleanarr = detect_cosmics(
            data + data_offset,
            inmask=None,
            sigclip=sigclip,
            sigfrac=sigfrac,
            objlim=objlim,
            gain=gain.value,
            readnoise=readnoise.value,
            satlevel=satlevel,
            niter=niter,
            sepmed=sepmed,
            cleantype=cleantype,
            fsmode=fsmode,
            psfmodel=psfmodel,
            psffwhm=psffwhm,
            psfsize=psfsize,
            psfk=psfk,
            psfbeta=psfbeta,
            verbose=verbose,
            **asy_background_kwargs,
        )

        cleanarr = cleanarr - data_offset
        cleanarr = _astroscrappy_gain_apply_helper(
            cleanarr, gain.value, gain_apply, old_astroscrappy_interface
        )

        return cleanarr, crmask

    elif isinstance(ccd, CCDData):
        # Start with a check for a special case: ccd is in electron, and
        # gain and readnoise have no units. In that case we issue a warning
        # instead of raising an error to avoid crashing user's pipelines.
        if ccd.unit.is_equivalent(u.electron) and gain.value != 1.0:
            warnings.warn(
                "Image unit is electron but gain value "
                "is not 1.0. Data maybe end up being gain "
                "corrected twice.",
                stacklevel=2,
            )

        else:
            if (
                (readnoise.unit == u.electron)
                and (ccd.unit == u.electron)
                and (gain.value == 1.0)
            ):
                gain = gain.value * u.one
            # Check unit consistency before taking the time to check for
            # cosmic rays.
            if not (gain * ccd).unit.is_equivalent(readnoise.unit):
                raise ValueError(
                    f"Inconsistent units for gain ({gain.unit}) "
                    + f" ccd ({ccd.unit}) and readnoise ({readnoise.unit})."
                )

        crmask, cleanarr = detect_cosmics(
            ccd.data + data_offset,
            inmask=ccd.mask,
            sigclip=sigclip,
            sigfrac=sigfrac,
            objlim=objlim,
            gain=gain.value,
            readnoise=readnoise.value,
            satlevel=satlevel,
            niter=niter,
            sepmed=sepmed,
            cleantype=cleantype,
            fsmode=fsmode,
            psfmodel=psfmodel,
            psffwhm=psffwhm,
            psfsize=psfsize,
            psfk=psfk,
            psfbeta=psfbeta,
            verbose=verbose,
            **asy_background_kwargs,
        )

        # create the new ccd data object
        nccd = ccd.copy()

        cleanarr = cleanarr - data_offset
        cleanarr = _astroscrappy_gain_apply_helper(
            cleanarr, gain.value, gain_apply, old_astroscrappy_interface
        )

        # Fix the units if the gain is being applied.
        nccd.unit = ccd.unit * gain.unit

        nccd.data = cleanarr
        if nccd.mask is None:
            nccd.mask = crmask
        else:
            nccd.mask = nccd.mask + crmask

        return nccd

    else:
        raise TypeError("ccd is not a CCDData or ndarray object.")


def _astroscrappy_gain_apply_helper(cleaned_data, gain, gain_apply, old_interface):
    """
    Helper function for logic determining how to apply gain to cleaned
    data. In the old astroscrappy interface cleaned data was always
    gain-corrected. In the new interface it is not. This function works out
    the Right Thing to do given the inputs.

    cleaned_data : `numpy.ndarray`
        The cleaned data.

    gain: float
        The gain to (maybe) be applied.

    gain_apply : bool
        If ``True``, the cleaned data should have the gain applied, otherwise
        the gain should be applied.

    old_interface : bool
        If ``True`` the old, i.e. pre-1.1.0, astroscrappy interface is being
        used. The old interface always gain corrected the cleaned data, the
        new one does not.
    """
    if gain != 1.0:
        if gain_apply:
            if not old_interface:
                # New interface does not gain correct, old one did.
                return cleaned_data * gain
        else:
            # Do not want gain correct
            if old_interface:  # pragma: no cover
                # Old interface gain corrected always so take it out
                return cleaned_data / gain
    return cleaned_data


def cosmicray_median(ccd, error_image=None, thresh=5, mbox=11, gbox=0, rbox=0):
    """
    Identify cosmic rays through median technique. The median technique
    identifies cosmic rays by identifying pixels by subtracting a median image
    from the initial data array.

    Parameters
    ----------
    ccd : `~astropy.nddata.CCDData`, `numpy.ndarray` or `numpy.ma.MaskedArray`
        Data to have cosmic ray cleaned.

    thresh : float, optional
        Threshold for detecting cosmic rays.
        Default is ``5``.

    error_image : `numpy.ndarray`, float or None, optional
        Error level. If None, the task will use the standard
        deviation of the data. If an ndarray, it should have the same shape
        as data.
        Default is ``None``.

    mbox : int, optional
        Median box for detecting cosmic rays.
        Default is ``11``.

    gbox : int, optional
        Box size to grow cosmic rays. If zero, no growing will be done.
        Default is ``0``.

    rbox : int, optional
        Median box for calculating replacement values. If zero, no pixels will
        be replaced.
        Default is ``0``.

    Notes
    -----
    Similar implementation to crmedian in iraf.imred.crutil.crmedian.

    Returns
    -------
    nccd : `~astropy.nddata.CCDData` or `numpy.ndarray`
        An object of the same type as ccd is returned. If it is a
        `~astropy.nddata.CCDData`, the mask attribute will also be updated with
        areas identified with cosmic rays masked.

    nccd : `numpy.ndarray`
        If an `numpy.ndarray` is provided as ccd, a boolean ndarray with the
        cosmic rays identified will also be returned.

    Examples
    --------
    1) Given an numpy.ndarray object, the syntax for running
       cosmicray_median would be:

       >>> newdata, mask = cosmicray_median(data, error_image=error,
       ...                                  thresh=5, mbox=11,
       ...                                  rbox=11, gbox=5)   # doctest: +SKIP

       where error is an array that is the same shape as data but
       includes the pixel error. This would return a data array, newdata,
       with the bad pixels replaced by the local median from a box of 11
       pixels; and it would return a mask indicating the bad pixels.

    2) Given an `~astropy.nddata.CCDData` object with an uncertainty frame, the syntax
       for running cosmicray_median would be:

       >>> newccd = cosmicray_median(ccd, thresh=5, mbox=11,
       ...                           rbox=11, gbox=5)   # doctest: +SKIP

       The newccd object will have bad pixels in its data array replace and the
       mask of the object will be created if it did not previously exist or be
       updated with the detected cosmic rays.
    """
    if isinstance(ccd, np.ndarray):
        data = ccd

        if error_image is None:
            error_image = data.std()
        else:
            if not isinstance(error_image, (float, np.ndarray)):
                raise TypeError("error_image is not a float or ndarray.")

        # create the median image
        marr = ndimage.median_filter(data, size=(mbox, mbox))

        # Only look at the data array
        if isinstance(data, np.ma.MaskedArray):
            data = data.data

        # Find the residual image
        rarr = (data - marr) / error_image

        # identify all sources
        crarr = rarr > thresh

        # grow the pixels
        if gbox > 0:
            crarr = ndimage.maximum_filter(crarr, gbox)

        # replace bad pixels in the image
        ndata = data.copy()
        if rbox > 0:
            data = np.ma.masked_array(data, (crarr == 1))
            mdata = ndimage.median_filter(data, rbox)
            ndata[crarr == 1] = mdata[crarr == 1]

        return ndata, crarr
    elif isinstance(ccd, CCDData):

        # set up the error image
        if error_image is None and ccd.uncertainty is not None:
            error_image = ccd.uncertainty.array
        if ccd.data.shape != error_image.shape:
            raise ValueError("error_image is not the same shape as data.")

        data, crarr = cosmicray_median(
            ccd.data,
            error_image=error_image,
            thresh=thresh,
            mbox=mbox,
            gbox=gbox,
            rbox=rbox,
        )

        # create the new ccd data object
        nccd = ccd.copy()
        nccd.data = data
        if nccd.mask is None:
            nccd.mask = crarr
        else:
            nccd.mask = nccd.mask + crarr
        return nccd

    else:
        raise TypeError("ccd is not an numpy.ndarray or a CCDData object.")


def ccdmask(
    ratio,
    findbadcolumns=False,
    byblocks=False,
    ncmed=7,
    nlmed=7,
    ncsig=15,
    nlsig=15,
    lsigma=9,
    hsigma=9,
    ngood=5,
):
    """
    Uses method based on the IRAF ccdmask task to generate a mask based on the
    given input.

    .. note::
        This function uses ``lines`` as synonym for the first axis and
        ``columns`` the second axis. Only two-dimensional ``ratio`` is
        currently supported.

    Parameters
    ----------
    ratio : `~astropy.nddata.CCDData`
        Data to used to form mask.  Typically this is the ratio of two flat
        field images.

    findbadcolumns : `bool`, optional
        If set to True, the code will search for bad column sections.  Note
        that this treats columns as special and breaks symmetry between lines
        and columns and so is likely only appropriate for detectors which have
        readout directions.
        Default is ``False``.

    byblocks : `bool`, optional
        If set to true, the code will divide the image up in to blocks of size
        nlsig by ncsig and determine the standard deviation estimate in each
        block (as described in the original IRAF task, see Notes below).  If
        set to False, then the code will use `scipy.ndimage.percentile_filter`
        to generate a running box version of the standard
        deviation estimate and use that value for the standard deviation at
        each pixel.
        Default is ``False``.

    ncmed, nlmed : `int`, optional
        The column and line size of the moving median rectangle used to
        estimate the uncontaminated local signal. The column median size should
        be at least 3 pixels to span single bad columns.
        Default is ``7``.

    ncsig, nlsig : `int`, optional
        The column and line size of regions used to estimate the uncontaminated
        local sigma using a percentile. The size of the box should contain of
        order 100 pixels or more.
        Default is ``15``.

    lsigma, hsigma : `float`, optional
        Positive sigma factors to use for selecting pixels below and above the
        median level based on the local percentile sigma.
        Default is ``9``.

    ngood : `int`, optional
        Gaps of undetected pixels along the column direction of length less
        than this amount are also flagged as bad pixels, if they are between
        pixels masked in that column.
        Default is ``5``.

    Returns
    -------
    mask : `numpy.ndarray`
        A boolean ndarray where the bad pixels have a value of 1 (True) and
        valid pixels 0 (False), following the numpy.ma conventions.

    Notes
    -----
    Similar implementation to IRAF's ccdmask task.
    The Following documentation is copied directly from:
    http://stsdas.stsci.edu/cgi-bin/gethelp.cgi?ccdmask

    The input image is first subtracted by a moving box median. The median is
    unaffected by bad pixels provided the median size is larger that twice the
    size of a bad region. Thus, if 3 pixel wide bad columns are present then
    the column median box size should be at least 7 pixels. The median box can
    be a single pixel wide along one dimension if needed. This may be
    appropriate for spectroscopic long slit data.

    The median subtracted image is then divided into blocks of size nclsig by
    nlsig. In each block the pixel values are sorted and the pixels nearest the
    30.9 and 69.1 percentile points are found; this would be the one sigma
    points in a Gaussian noise distribution. The difference between the two
    count levels divided by two is then the local sigma estimate. This
    algorithm is used to avoid contamination by the bad pixel values. The block
    size must be at least 10 pixels in each dimension to provide sufficient
    pixels for a good estimate of the percentile sigma. The sigma uncertainty
    estimate of each pixel in the image is then the sigma from the nearest
    block.

    The deviant pixels are found by comparing the median subtracted residual to
    a specified sigma threshold factor times the local sigma above and below
    zero (the lsigma and hsigma parameters). This is done for individual pixels
    and then for column sums of pixels (excluding previously flagged bad
    pixels) from two to the number of lines in the image. The sigma of the sums
    is scaled by the square root of the number of pixels summed so that
    statistically low or high column regions may be detected even though
    individual pixels may not be statistically deviant. For the purpose of this
    task one would normally select large sigma threshold factors such as six or
    greater to detect only true bad pixels and not the extremes of the noise
    distribution.

    As a final step each column is examined to see if there are small segments
    of unflagged pixels between bad pixels. If the length of a segment is less
    than that given by the ngood parameter all the pixels in the segment are
    also marked as bad.
    """
    try:
        nlines, ncols = ratio.data.shape
    except (TypeError, ValueError) as err:
        # shape is not iterable or has more or less than two values
        raise ValueError('"ratio" must be two-dimensional.') from err
    except AttributeError as err:
        # No data attribute or data has no shape attribute.
        raise ValueError('"ratio" should be a "CCDData".') from err

    def _sigma_mask(baseline, one_sigma_value, lower_sigma, upper_sigma):
        """Helper function to mask values outside of the specified sigma range."""
        return (baseline < -lower_sigma * one_sigma_value) | (
            baseline > upper_sigma * one_sigma_value
        )

    mask = ~np.isfinite(ratio.data)
    medsub = ratio.data - ndimage.median_filter(ratio.data, size=(nlmed, ncmed))

    if byblocks:
        nlinesblock = int(math.ceil(nlines / nlsig))
        ncolsblock = int(math.ceil(ncols / ncsig))
        for i in range(nlinesblock):
            for j in range(ncolsblock):
                l1 = i * nlsig
                l2 = min((i + 1) * nlsig, nlines)
                c1 = j * ncsig
                c2 = min((j + 1) * ncsig, ncols)
                block = medsub[l1:l2, c1:c2]
                high = np.percentile(block.ravel(), 69.1)
                low = np.percentile(block.ravel(), 30.9)
                block_sigma = (high - low) / 2.0
                block_mask = _sigma_mask(block, block_sigma, lsigma, hsigma)
                mblock = np.ma.MaskedArray(block, mask=block_mask, copy=False)

                if findbadcolumns:
                    csum = np.ma.sum(mblock, axis=0)
                    csum[csum <= 0] = 0
                    csum_sigma = np.ma.MaskedArray(np.sqrt(c2 - c1 - csum))
                    colmask = _sigma_mask(csum.filled(1), csum_sigma, lsigma, hsigma)
                    block_mask[:, :] |= colmask[np.newaxis, :]

                mask[l1:l2, c1:c2] = block_mask
    else:
        high = ndimage.percentile_filter(medsub, 69.1, size=(nlsig, ncsig))
        low = ndimage.percentile_filter(medsub, 30.9, size=(nlsig, ncsig))
        sigmas = (high - low) / 2.0
        mask |= _sigma_mask(medsub, sigmas, lsigma, hsigma)

    if findbadcolumns:
        # Loop through columns and look for short segments (<ngood pixels long)
        # which are unmasked, but are surrounded by masked pixels and mask them
        # under the assumption that the column region is bad.
        for col in range(0, ncols):
            for line in range(0, nlines - ngood - 1):
                if mask[line, col]:
                    for i in range(2, ngood + 2):
                        lend = line + i
                        if mask[lend, col] and not np.all(mask[line : lend + 1, col]):
                            mask[line:lend, col] = True
    return mask


def bitfield_to_boolean_mask(bitfield, ignore_bits=0, flip_bits=None):
    """Convert an integer bit field to a boolean mask.

    Parameters
    ----------
    bitfield : `numpy.ndarray` of integer dtype
        The array of bit flags.

    ignore_bits : int, None or str, optional
        The bits to ignore when converting the bitfield.

        - If it's an integer it's binary representation is interpreted as the
          bits to ignore. ``0`` means that all bit flags are taken into
          account while a binary representation of all ``1`` means that all
          flags would be ignored.
        - If it's ``None`` then all flags are ignored
        - If it's a string then it must be a ``,`` or ``+`` separated string
          of integers that bits to ignore. If the string starts with an ``~``
          the integers are interpreted as **the only flags** to take into
          account.

        Default is ``0``.

    Returns
    -------
    mask : `numpy.ndarray` of boolean dtype
        The bitfield converted to a boolean mask that can be used for
        `numpy.ma.MaskedArray` or `~astropy.nddata.CCDData`.

    Examples
    --------
    Bitfields (or data quality arrays) are integer arrays where the binary
    representation of the values indicates whether a specific flag is set or
    not. The convention is that a value of ``0`` represents a **good value**
    and a value that is ``!= 0`` represents a value that is in some (or
    multiple) ways considered a **bad value**. The ``bitfield_to_boolean_mask``
    function can be used by default to create a boolean mask wherever any bit
    flag is set::

        >>> import ccdproc
        >>> import numpy as np
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8))
        array([False,  True,  True,  True,  True,  True,  True,  True]...)

    To ignore all bit flags ``ignore_bits=None`` can be used::

        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits=None)
        array([False, False, False, False, False, False, False, False]...)

    To ignore only specific bit flags one can use a ``list`` of bits flags to
    ignore::

        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits=[1, 4])
        array([False, False,  True,  True, False, False,  True,  True]...)

    There are some equivalent ways::

        >>> # pass in the sum of the "ignore_bits" directly
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits=5)  # 1 + 4
        array([False, False,  True,  True, False, False,  True,  True]...)
        >>> # use a comma seperated string of integers
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits='1, 4')
        array([False, False,  True,  True, False, False,  True,  True]...)
        >>> # use a + seperated string of integers
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits='1+4')
        array([False, False,  True,  True, False, False,  True,  True]...)

    Instead of directly specifying the **bits flags to ignore** one can also
    pass in the **only bits that shouldn't be ignored** by prepending a ``~``
    to the string of ``ignore_bits`` (or if it's not a string in
    ``ignore_bits`` one can set ``flip_bits=True``)::

        >>> # ignore all bit flags except the one for 2.
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits='~(2)')
        array([False, False,  True,  True, False, False,  True,  True]...)
        >>> # ignore all bit flags except the one for 1, 8 and 32.
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits='~(1, 8, 32)')
        array([False,  True, False,  True, False,  True, False,  True]...)

        >>> # Equivalent for a list using flip_bits.
        >>> ccdproc.bitfield_to_boolean_mask(np.arange(8), ignore_bits=[1, 8, 32],
        ...                                  flip_bits=True)
        array([False,  True, False,  True, False,  True, False,  True]...)

    """
    return _bitfield_to_boolean_mask(
        bitfield, ignore_bits, flip_bits=flip_bits, good_mask_value=False, dtype=bool
    )


class Keyword:
    """ """

    def __init__(self, name, unit=None, value=None):
        self._name = name
        self._unit = unit
        self.value = value

    @property
    def name(self):
        return self._name

    @property
    def unit(self):
        return self._unit

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value is None:
            self._value = value
        elif isinstance(value, Quantity):
            self._unit = value.unit
            self._value = value
        elif isinstance(value, str):
            if self.unit is not None:
                raise ValueError("keyword with a unit cannot have a " "string value.")
            else:
                self._value = value
        else:
            if self.unit is None:
                raise ValueError(
                    "no unit provided. Set value with " "an astropy.units.Quantity."
                )
            self._value = value * self.unit

    def value_from(self, header):
        """
        Set value of keyword from FITS header.

        Parameters
        ----------
        header : `~astropy.io.fits.Header`
            FITS header containing a value for this keyword.
        """
        value_from_header = header[self.name]
        self.value = value_from_header
        return self.value



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from astropy.nddata import CCDData

from ccdproc import Keyword, create_deviation, subtract_bias, trim_image
from ccdproc.core import _short_names
from ccdproc.tests.pytest_fixtures import ccd_data as ccd_data_func


@pytest.mark.parametrize("key", ["short", "toolongforfits"])
def test_log_string(key):
    ccd_data = ccd_data_func()
    add_key = key
    new = create_deviation(ccd_data, readnoise=3 * ccd_data.unit, add_keyword=add_key)
    # Keys should be added to new but not to ccd_data and should have
    # no value.
    assert add_key in new.meta
    assert add_key not in ccd_data.meta
    # Long keyword names should be accessible with just the keyword name
    # without HIERARCH -- is it?
    assert new.meta[add_key] is None


def test_log_keyword():
    ccd_data = ccd_data_func()
    key = "filter"
    key_val = "V"
    kwd = Keyword(key, value=key_val)
    new = create_deviation(ccd_data, readnoise=3 * ccd_data.unit, add_keyword=kwd)
    # Was the Keyword added with the correct value?
    assert kwd.name in new.meta
    assert kwd.name not in ccd_data.meta
    assert new.meta[kwd.name] == key_val


def test_log_dict():
    ccd_data = ccd_data_func()
    keys_to_add = {
        "process": "Added deviation",
        "n_images_input": 1,
        "current_temp": 42.9,
    }
    new = create_deviation(
        ccd_data, readnoise=3 * ccd_data.unit, add_keyword=keys_to_add
    )
    for k, v in keys_to_add.items():
        # Were all dictionary items added?
        assert k in new.meta
        assert k not in ccd_data.meta
        assert new.meta[k] == v


def test_log_bad_type_fails():
    ccd_data = ccd_data_func()
    add_key = 15  # anything not string and not dict-like will work here
    # Do we fail with non-string, non-Keyword, non-dict-like value?
    with pytest.raises(AttributeError):
        create_deviation(ccd_data, readnoise=3 * ccd_data.unit, add_keyword=add_key)


def test_log_set_to_None_does_not_change_header():
    ccd_data = ccd_data_func()
    new = create_deviation(ccd_data, readnoise=3 * ccd_data.unit, add_keyword=None)
    assert new.meta.keys() == ccd_data.header.keys()


def test_implicit_logging():
    ccd_data = ccd_data_func()
    # If nothing is supplied for the add_keyword argument then the following
    # should happen:
    # + A key named func.__name__ is created, with
    # + value that is the list of arguments the function was called with.
    bias = CCDData(np.zeros_like(ccd_data.data), unit="adu")
    result = subtract_bias(ccd_data, bias)
    assert "subtract_bias" in result.header
    assert result.header["subtract_bias"] == (
        "subbias",
        "Shortened name for ccdproc command",
    )
    assert result.header["subbias"] == "ccd=<CCDData>, master=<CCDData>"

    result = create_deviation(ccd_data, readnoise=3 * ccd_data.unit)
    assert result.header["create_deviation"] == (
        "creatvar",
        "Shortened name for ccdproc command",
    )
    assert "readnoise=" + str(3 * ccd_data.unit) in result.header["creatvar"]


def test_loggin_without_keyword_args():
    # Regression test for the first failure in #704, which fails because
    # there is no "fits_section" keyword in the call to trim_image.
    ccd = CCDData(data=np.arange(1000).reshape(20, 50), header=None, unit="adu")
    section = "[10:20, 10:20]"
    trim_1 = trim_image(ccd, "[10:20, 10:20]")
    assert section in trim_1.header[_short_names["trim_image"]]


def test_logging_with_really_long_parameter_value():
    # Another regression test for the trim_3 case in #704
    ccd = CCDData(data=np.arange(1000).reshape(20, 50), header=None, unit="adu")
    section = (
        "[10:2000000000000000000000000000000000000000000000000000000, "
        "10:2000000000000000000000000000000]"
    )
    trim_3 = trim_image(ccd, fits_section=section)
    assert section in trim_3.header[_short_names["trim_image"]]



import pytest

pytest.importorskip("memory_profiler")

import gc
import sys
from argparse import ArgumentParser
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import psutil
from astropy.io import fits
from astropy.nddata import CCDData
from astropy.stats import median_absolute_deviation
from memory_profiler import memory_usage

# This bit of hackery ensures that we can see ccdproc from within
# the test suite
sys.path.append(str(Path().cwd()))
from ccdproc import ImageFileCollection, combine
from ccdproc.combiner import _calculate_size_of_image

# Do not combine these into one statement. When all references are lost
# to a TemporaryDirectory the directory is automatically deleted. _TMPDIR
# creates a reference that will stick around.
_TMPDIR = TemporaryDirectory()
TMPPATH = Path(_TMPDIR.name)


def generate_fits_files(n_images, size=None, seed=1523):
    if size is None:
        use_size = (2024, 2031)
    else:
        use_size = (size, size)

    rng = np.random.default_rng(seed=seed)

    base_name = "test-combine-{num:03d}.fits"

    for num in range(n_images):
        data = rng.normal(size=use_size)
        # Now add some outlying pixels so there is something to clip
        n_bad = 50000
        bad_x = rng.integers(0, high=use_size[0] - 1, size=n_bad)
        bad_y = rng.integers(0, high=use_size[1] - 1, size=n_bad)
        data[bad_x, bad_y] = rng.choice([-1, 1], size=n_bad) * (10 + rng.random(n_bad))
        hdu = fits.PrimaryHDU(data=np.asarray(data, dtype="float32"))
        hdu.header["for_prof"] = "yes"
        hdu.header["bunit"] = "adu"
        path = TMPPATH.resolve() / base_name.format(num=num)
        hdu.writeto(path, overwrite=True)


def run_memory_profile(
    n_files,
    sampling_interval,
    sigma_clip=False,
    combine_method=None,
    memory_limit=None,
):
    """
    Try opening a bunch of files with a relatively low limit on the number
    of open files.

    Parameters
    ----------

    n_files : int
        Number of files to combine.

    sampling_interval : float
        Time, in seconds, between memory samples.

    size : int, optional
        Size of one side of the image (the image is always square).

    sigma_clip : bool, optional
        If true, sigma clip the data before combining.

    combine_method : str, optional
        Should be one of the combine methods accepted by
        ccdproc.combine

    memory_limit : int, optional
        Cap on memory use during image combination.
    """
    # Do a little input validation
    if n_files <= 0:
        raise ValueError("Argument 'n' must be a positive integer")

    proc = psutil.Process()

    print("Process ID is: ", proc.pid, flush=True)
    ic = ImageFileCollection(str(TMPPATH))
    files = ic.files_filtered(for_prof="yes", include_path=True)

    kwargs = {"method": combine_method}

    if sigma_clip:
        kwargs.update(
            {
                "sigma_clip": True,
                "sigma_clip_low_thresh": 5,
                "sigma_clip_high_thresh": 5,
                "sigma_clip_func": np.ma.median,
                "sigma_clip_dev_func": median_absolute_deviation,
            }
        )

    ccd = CCDData.read(files[0])
    expected_img_size = _calculate_size_of_image(ccd, None)

    if memory_limit:
        kwargs["mem_limit"] = memory_limit

    pre_mem_use = memory_usage(-1, interval=sampling_interval, timeout=1)
    baseline = np.mean(pre_mem_use)
    print(f"Subtracting baseline memory before profile: {baseline}")
    mem_use = memory_usage(
        (combine, (files,), kwargs), interval=sampling_interval, timeout=None
    )
    mem_use = [m - baseline for m in mem_use]
    return mem_use, expected_img_size


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("number", type=int, help="Number of files to combine.")
    parser.add_argument(
        "--size",
        type=int,
        action="store",
        help="Size of one side of image to create. "
        "All images are square, so only give "
        "a single number for the size.",
    )
    parser.add_argument(
        "--combine-method",
        "-c",
        choices=("average", "median"),
        help="Method to use to combine images.",
    )
    parser.add_argument(
        "--memory-limit", type=int, help="Limit combination to this amount of memory"
    )
    parser.add_argument(
        "--sigma-clip",
        action="store_true",
        help="If set, sigma-clip before combining. Clipping "
        "will be done with high/low limit of 5. "
        "The central function is the median, the "
        "deviation is the median_absolute_deviation.",
    )
    parser.add_argument(
        "--sampling-freq",
        type=float,
        default=0.05,
        help="Time, in seconds, between memory samples.",
    )
    parser.add_argument(
        "--frequent-gc",
        action="store_true",
        help="If set, perform garbage collection "
        "much more frequently than the default.",
    )
    args = parser.parse_args()

    if args.frequent_gc:
        gc.set_threshold(10, 10, 10)

    print("Garbage collection thresholds: ", gc.get_threshold())

    mem_use = run_memory_profile(
        args.number,
        args.sampling_freq,
        sigma_clip=args.sigma_clip,
        combine_method=args.combine_method,
        memory_limit=args.memory_limit,
    )
    print("Max memory usage (MB): ", np.max(mem_use))
    print("Baseline memory usage (MB): ", mem_use[0])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from sys import platform

import numpy as np
import pytest

try:
    from ccdproc.tests.run_for_memory_profile import (
        TMPPATH,
        generate_fits_files,
        run_memory_profile,
    )
except ImportError:
    memory_profile_present = False
else:
    memory_profile_present = True

image_size = 2000  # Square image, so 4000 x 4000
num_files = 10


def setup_module():
    if memory_profile_present:
        generate_fits_files(num_files, size=image_size)


def teardown_module():
    if memory_profile_present:
        for fil in TMPPATH.glob("*.fit"):
            fil.unlink()


@pytest.mark.skipif(
    not platform.startswith("linux"), reason="memory tests only work on linux"
)
@pytest.mark.skipif(not memory_profile_present, reason="memory_profiler not installed")
@pytest.mark.parametrize("combine_method", ["average", "sum", "median"])
def test_memory_use_in_combine(combine_method):
    # This is essentially a regression test for
    # https://github.com/astropy/ccdproc/issues/638
    #
    sampling_interval = 0.01  # sec
    memory_limit = 500000000  # bytes, roughly 0.5GB

    mem_use, _ = run_memory_profile(
        num_files,
        sampling_interval,
        memory_limit=memory_limit,
        combine_method=combine_method,
    )

    mem_use = np.array(mem_use)
    # We do not expect memory use to be strictly less than memory_limit
    # throughout the combination. The factor below allows for that.
    # It may need to be raised in the future...that is fine, there is a
    # separate test for average memory use.
    overhead_allowance = 1.75

    # memory_profile reports in MB (no, this is not the correct conversion)
    memory_limit_mb = memory_limit / 1e6

    # Checks for TOO MUCH MEMORY USED

    # Check peak memory use
    assert np.max(mem_use) <= overhead_allowance * memory_limit_mb

    # Also check average, which gets no allowance
    assert np.mean(mem_use) < memory_limit_mb

    # Checks for NOT ENOUGH MEMORY USED; if these fail it means that
    # memory_factor in the combine function should perhaps be modified

    # DROPPED THESE TESTS -- it isn't clear they were actually useful and
    # in any event the important thing to guarantee is that we don't
    # exceed the memory limit.

    # If the peak is coming in under the limit something need to be fixed
    # assert np.max(mem_use) >= 0.95 * memory_limit_mb

    # If the average is really low perhaps we should look at reducing peak
    # usage. Nothing special, really, about the factor 0.4 below.
    # assert np.mean(mem_use[mem_use > 0]) > 0.4 * memory_limit_mb



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import inspect
from functools import wraps
from itertools import chain

import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData

import ccdproc  # Really only need Keyword from ccdproc

__all__ = []

_LOG_ARGUMENT = "add_keyword"

_LOG_ARG_HELP = f"""
    {_LOG_ARGUMENT} : str, `~ccdproc.Keyword` or dict-like, optional
        Item(s) to add to metadata of result. Set to False or None to
        completely  disable logging.
        Default is to add a dictionary with a single item:
        The key is the name of this function  and the value is a string
        containing the arguments the function was called with, except the
        value of this argument.
    """


def _insert_in_metadata_fits_safe(ccd, key, value):
    from .core import _short_names

    if key in _short_names:
        # This keyword was (hopefully) added by autologging but the
        # combination of it and its value not FITS-compliant in two
        # ways: the keyword name may be more than 8 characters and
        # the value may be too long. FITS cannot handle both of
        # those problems at once, so this fixes one of those
        # problems...
        # Shorten, sort of...
        short_name = _short_names[key]
        if isinstance(ccd.meta, fits.Header):
            ccd.meta[f"HIERARCH {key.upper()}"] = (
                short_name,
                "Shortened name for ccdproc command",
            )
        else:
            ccd.meta[key] = (short_name, "Shortened name for ccdproc command")
        ccd.meta[short_name] = value
    else:
        ccd.meta[key] = value


def log_to_metadata(func):
    """
    Decorator that adds logging to ccdproc functions.

    The decorator adds the optional argument _LOG_ARGUMENT to function
    signature and updates the function's docstring to reflect that.

    It also sets the default value of the argument to the name of the function
    and the arguments it was called with.
    """
    func.__doc__ = func.__doc__.format(log=_LOG_ARG_HELP)

    argspec = inspect.getfullargspec(func)
    original_args, _, _, defaults = (
        argspec.args,
        argspec.varargs,
        argspec.varkw,
        argspec.defaults,
    )

    # Add logging keyword and its default value for docstring
    original_args.append(_LOG_ARGUMENT)
    try:
        defaults = list(defaults)
    except TypeError:
        defaults = []
    defaults.append(True)

    signature_with_arg_added = inspect.signature(func)
    signature_with_arg_added = f"{func.__name__}{signature_with_arg_added}"
    func.__doc__ = "\n".join([signature_with_arg_added, func.__doc__])

    @wraps(func)
    def wrapper(*args, **kwd):
        # Grab the logging keyword, if it is present.
        log_result = kwd.pop(_LOG_ARGUMENT, True)
        result = func(*args, **kwd)

        if not log_result:
            # No need to add metadata....
            meta_dict = {}
        elif log_result is not True:
            meta_dict = _metadata_to_dict(log_result)
        else:
            # Logging is not turned off, but user did not provide a value
            # so construct one unless the config parameter auto_logging is set to False
            if ccdproc.conf.auto_logging:
                key = func.__name__
                # Get names of arguments, which may or may not have
                # been called as keywords.
                positional_args = original_args[: len(args)]

                all_args = chain(zip(positional_args, args), kwd.items())
                all_args = [
                    f"{name}={_replace_array_with_placeholder(val)}"
                    for name, val in all_args
                ]
                log_val = ", ".join(all_args)
                log_val = log_val.replace("\n", "")
                meta_dict = {key: log_val}
            else:
                meta_dict = {}

        for k, v in meta_dict.items():
            _insert_in_metadata_fits_safe(result, k, v)
        return result

    return wrapper


def _metadata_to_dict(arg):
    if isinstance(arg, str):
        # add the key, no value
        return {arg: None}
    elif isinstance(arg, ccdproc.Keyword):
        return {arg.name: arg.value}
    else:
        return arg


def _replace_array_with_placeholder(value):
    return_type_not_value = False
    if isinstance(value, u.Quantity):
        return_type_not_value = not value.isscalar
    elif isinstance(value, (NDData, np.ndarray)):
        try:
            length = len(value)
        except TypeError:
            # Value has no length...
            try:
                # ...but if it is NDData its .data will have a length
                length = len(value.data)
            except TypeError:
                # No idea what this data is, assume length is not 1
                length = 42
        return_type_not_value = length > 1

    if return_type_not_value:
        return f"<{value.__class__.__name__}>"
    else:
        return value



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.utils.exceptions import AstropyDeprecationWarning

from ccdproc.core import rebin
from ccdproc.tests.pytest_fixtures import ccd_data as ccd_data_func


# test rebinning ndarray
def test_rebin_ndarray():
    with pytest.raises(TypeError), pytest.warns(AstropyDeprecationWarning):
        rebin(1, (5, 5))


# test rebinning dimensions
def test_rebin_dimensions():
    ccd_data = ccd_data_func(data_size=10)
    with pytest.raises(ValueError), pytest.warns(AstropyDeprecationWarning):
        rebin(ccd_data.data, (5,))


# test rebinning dimensions
def test_rebin_ccddata_dimensions():
    ccd_data = ccd_data_func(data_size=10)
    with pytest.raises(ValueError), pytest.warns(AstropyDeprecationWarning):
        rebin(ccd_data, (5,))


# test rebinning works
def test_rebin_larger():
    ccd_data = ccd_data_func(data_size=10)
    a = ccd_data.data
    with pytest.warns(AstropyDeprecationWarning):
        b = rebin(a, (20, 20))

    assert b.shape == (20, 20)
    np.testing.assert_almost_equal(b.sum(), 4 * a.sum())


# test rebinning is invariant
def test_rebin_smaller():
    ccd_data = ccd_data_func(data_size=10)
    a = ccd_data.data
    with pytest.warns(AstropyDeprecationWarning):
        b = rebin(a, (20, 20))
        c = rebin(b, (10, 10))

    assert c.shape == (10, 10)
    assert (c - a).sum() == 0


# test rebinning with ccddata object
@pytest.mark.parametrize("mask_data, uncertainty", [(False, False), (True, True)])
def test_rebin_ccddata(mask_data, uncertainty):
    ccd_data = ccd_data_func(data_size=10)
    if mask_data:
        ccd_data.mask = np.zeros_like(ccd_data)
    if uncertainty:
        err = np.random.default_rng().normal(size=ccd_data.shape)
        ccd_data.uncertainty = StdDevUncertainty(err)

    with pytest.warns(AstropyDeprecationWarning):
        b = rebin(ccd_data, (20, 20))

    assert b.shape == (20, 20)
    if mask_data:
        assert b.mask.shape == (20, 20)
    if uncertainty:
        assert b.uncertainty.array.shape == (20, 20)


def test_rebin_does_not_change_input():
    ccd_data = ccd_data_func()
    original = ccd_data.copy()
    with pytest.warns(AstropyDeprecationWarning):
        _ = rebin(ccd_data, (20, 20))
    np.testing.assert_allclose(original.data, ccd_data.data)
    assert original.unit == ccd_data.unit



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from shutil import rmtree

import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import CCDData

from ..utils.sample_directory import directory_for_testing

# If additional pytest markers are defined the key in the dictionary below
# should be the name of the marker.
DEFAULTS = {"seed": 123, "data_size": 100, "data_scale": 1.0, "data_mean": 0.0}

DEFAULT_SEED = 123
DEFAULT_DATA_SIZE = 100
DEFAULT_DATA_SCALE = 1.0
DEFAULT_DATA_MEAN = 0.0


def value_from_markers(key, request):
    m = request.node.get_closest_marker(key)
    if m is not None:
        return m.args[0]
    else:
        return DEFAULTS[key]


def ccd_data(
    data_size=DEFAULT_DATA_SIZE,
    data_scale=DEFAULT_DATA_SCALE,
    data_mean=DEFAULT_DATA_MEAN,
    rng_seed=DEFAULT_SEED,
):
    """
    Return a CCDData object with units of ADU.

    The size of the data array is 100x100 but can be changed using the marker
    @pytest.mark.data_size(N) on the test function, where N should be the
    desired dimension.

    Data values are initialized to random numbers drawn from a normal
    distribution with mean of 0 and scale 1.

    The scale can be changed with the marker @pytest.marker.scale(s) on the
    test function, where s is the desired scale.

    The mean can be changed with the marker @pytest.marker.scale(m) on the
    test function, where m is the desired mean.
    """
    size = data_size
    scale = data_scale
    mean = data_mean

    ##Create random number generator with a specified state
    rng = np.random.default_rng(seed=rng_seed)

    data = rng.normal(loc=mean, size=[size, size], scale=scale)

    fake_meta = {"my_key": 42, "your_key": "not 42"}
    ccd = CCDData(data, unit=u.adu)
    ccd.header = fake_meta
    return ccd


@pytest.fixture
def triage_setup(request):

    n_test, test_dir = directory_for_testing()

    def teardown():
        try:
            rmtree(test_dir)
        except OSError:
            # If we cannot clean up just keep going.
            pass

    request.addfinalizer(teardown)

    class Result:
        def __init__(self, n, directory):
            self.n_test = n
            self.test_dir = directory

    return Result(n_test, test_dir)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy.units as u
import numpy as np
import pytest

from ccdproc.core import Keyword, create_deviation, gain_correct
from ccdproc.tests.pytest_fixtures import ccd_data as ccd_data_func


# tests for gain
@pytest.mark.parametrize(
    "gain",
    [
        3.0,
        3.0 * u.photon / u.adu,
        3.0 * u.electron / u.adu,
        Keyword("gainval", unit=u.electron / u.adu),
    ],
)
def test_linear_gain_correct(gain):
    ccd_data = ccd_data_func()
    # The data values should be positive, so the poisson noise calculation
    # works without throwing warnings
    ccd_data.data = np.absolute(ccd_data.data)
    ccd_data = create_deviation(ccd_data, readnoise=1.0 * u.adu)
    ccd_data.meta["gainval"] = 3.0
    orig_data = ccd_data.data
    ccd = gain_correct(ccd_data, gain)
    if isinstance(gain, Keyword):
        gain = gain.value  # convert to Quantity...
    try:
        gain_value = gain.value
    except AttributeError:
        gain_value = gain

    np.testing.assert_array_almost_equal_nulp(ccd.data, gain_value * orig_data)
    np.testing.assert_array_almost_equal_nulp(
        ccd.uncertainty.array, gain_value * ccd_data.uncertainty.array
    )

    if isinstance(gain, u.Quantity):
        assert ccd.unit == ccd_data.unit * gain.unit
    else:
        assert ccd.unit == ccd_data.unit


# test gain with gain_unit
def test_linear_gain_unit_keyword():
    ccd_data = ccd_data_func()
    # The data values should be positive, so the poisson noise calculation
    # works without throwing warnings
    ccd_data.data = np.absolute(ccd_data.data)

    ccd_data = create_deviation(ccd_data, readnoise=1.0 * u.adu)
    orig_data = ccd_data.data
    gain = 3.0
    gain_unit = u.electron / u.adu
    ccd = gain_correct(ccd_data, gain, gain_unit=gain_unit)
    np.testing.assert_array_almost_equal_nulp(ccd.data, gain * orig_data)
    np.testing.assert_array_almost_equal_nulp(
        ccd.uncertainty.array, gain * ccd_data.uncertainty.array
    )
    assert ccd.unit == ccd_data.unit * gain_unit



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest
from astropy import units as u
from astropy.io import fits

from ccdproc.core import Keyword


def test_keyword_init():
    key_name = "some_key"
    key = Keyword(key_name, unit=u.second)
    assert key.name == key_name
    assert key.unit == u.second


def test_keyword_properties_read_only():
    key = Keyword("observer")
    with pytest.raises(AttributeError):
        key.name = "error"
    with pytest.raises(AttributeError):
        key.unit = u.hour


unit = u.second
numerical_value = 30


# The variable "expected" below is
#     True if the expected result is key.value == numerical_value * key.unit
#     Name of an error if an error is expected
#     A string if the expected value is a string
@pytest.mark.parametrize(
    "value,unit,expected",
    [
        (numerical_value, unit, True),
        (numerical_value, None, ValueError),
        (numerical_value * unit, None, True),
        (numerical_value * unit, unit, True),
        (numerical_value * unit, u.km, True),
        ("some string", None, "some string"),
        ("no strings with unit", unit, ValueError),
    ],
)
def test_value_setting(value, unit, expected):
    name = "exposure"
    # Setting at initialization time with
    try:
        expected_is_error = issubclass(expected, Exception)
    except TypeError:
        expected_is_error = False
    if expected_is_error:
        with pytest.raises(expected):
            key = Keyword(name, unit=unit, value=value)
    else:
        key = Keyword(name, unit=unit, value=value)
        if isinstance(expected, str):
            assert key.value == expected
        else:
            assert key.value == numerical_value * key.unit


def test_keyword_value_from_header():
    name = "exposure"
    numerical_value = 30
    unit = u.second
    h = fits.Header()
    h[name] = numerical_value

    key = Keyword(name, unit=unit)
    assert key.value_from(h) == numerical_value * unit
    assert key.value == numerical_value * unit



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This module implements the combiner class."""

import numpy as np
from numpy import ma

try:
    import bottleneck as bn
except ImportError:
    HAS_BOTTLENECK = False  # pragma: no cover
else:
    HAS_BOTTLENECK = True

from astropy import log
from astropy.nddata import CCDData, StdDevUncertainty
from astropy.stats import sigma_clip
from astropy.utils import deprecated_renamed_argument

from .core import sigma_func

__all__ = ["Combiner", "combine"]


def _default_median():  # pragma: no cover
    if HAS_BOTTLENECK:
        return bn.nanmedian
    else:
        return np.nanmedian


def _default_average():  # pragma: no cover
    if HAS_BOTTLENECK:
        return bn.nanmean
    else:
        return np.nanmean


def _default_sum():  # pragma: no cover
    if HAS_BOTTLENECK:
        return bn.nansum
    else:
        return np.nansum


def _default_std():  # pragma: no cover
    if HAS_BOTTLENECK:
        return bn.nanstd
    else:
        return np.nanstd


_default_sum_func = _default_sum()
_default_std_dev_func = _default_std()


class Combiner:
    """
    A class for combining CCDData objects.

    The Combiner class is used to combine together `~astropy.nddata.CCDData` objects
    including the method for combining the data, rejecting outlying data,
    and weighting used for combining frames.

    Parameters
    ----------
    ccd_iter : list or generator
        A list or generator of CCDData objects that will be combined together.

    dtype : str or `numpy.dtype` or None, optional
        Allows user to set dtype. See `numpy.array` ``dtype`` parameter
        description. If ``None`` it uses ``np.float64``.
        Default is ``None``.

    Raises
    ------
    TypeError
        If the ``ccd_iter`` are not `~astropy.nddata.CCDData` objects, have different
        units, or are different shapes.

    Examples
    --------
    The following is an example of combining together different
    `~astropy.nddata.CCDData` objects::

        >>> import numpy as np
        >>> import astropy.units as u
        >>> from astropy.nddata import CCDData
        >>> from ccdproc import Combiner
        >>> ccddata1 = CCDData(np.ones((4, 4)), unit=u.adu)
        >>> ccddata2 = CCDData(np.zeros((4, 4)), unit=u.adu)
        >>> ccddata3 = CCDData(np.ones((4, 4)), unit=u.adu)
        >>> c = Combiner([ccddata1, ccddata2, ccddata3])
        >>> ccdall = c.average_combine()
        >>> ccdall  # doctest: +FLOAT_CMP
        CCDData([[ 0.66666667,  0.66666667,  0.66666667,  0.66666667],
                 [ 0.66666667,  0.66666667,  0.66666667,  0.66666667],
                 [ 0.66666667,  0.66666667,  0.66666667,  0.66666667],
                 [ 0.66666667,  0.66666667,  0.66666667,  0.66666667]]...)
    """

    def __init__(self, ccd_iter, dtype=None):
        if ccd_iter is None:
            raise TypeError(
                "ccd_iter should be a list or a generator of CCDData objects."
            )

        if dtype is None:
            dtype = np.float64

        default_shape = None
        default_unit = None

        ccd_list = list(ccd_iter)

        for ccd in ccd_list:
            # raise an error if the objects aren't CCDData objects
            if not isinstance(ccd, CCDData):
                raise TypeError("ccd_list should only contain CCDData objects.")

            # raise an error if the shape is different
            if default_shape is None:
                default_shape = ccd.shape
            else:
                if not (default_shape == ccd.shape):
                    raise TypeError("CCDData objects are not the same size.")

            # raise an error if the units are different
            if default_unit is None:
                default_unit = ccd.unit
            else:
                if not (default_unit == ccd.unit):
                    raise TypeError("CCDData objects don't have the same unit.")

        self.ccd_list = ccd_list
        self.unit = default_unit
        self.weights = None
        self._dtype = dtype

        # set up the data array
        new_shape = (len(ccd_list),) + default_shape
        self.data_arr = ma.masked_all(new_shape, dtype=dtype)

        # populate self.data_arr
        for i, ccd in enumerate(ccd_list):
            self.data_arr[i] = ccd.data
            if ccd.mask is not None:
                self.data_arr.mask[i] = ccd.mask
            else:
                self.data_arr.mask[i] = ma.zeros(default_shape)

        # Must be after self.data_arr is defined because it checks the
        # length of the data array.
        self.scaling = None

    @property
    def dtype(self):
        return self._dtype

    @property
    def weights(self):
        """
        Weights used when combining the `~astropy.nddata.CCDData` objects.

        Parameters
        ----------
        weight_values : `numpy.ndarray` or None
            An array with the weight values. The dimensions should match the
            the dimensions of the data arrays being combined.
        """
        return self._weights

    @weights.setter
    def weights(self, value):
        if value is not None:
            if isinstance(value, np.ndarray):
                if value.shape != self.data_arr.data.shape:
                    if value.ndim != 1:
                        raise ValueError(
                            "1D weights expected when shapes of the "
                            "data and weights differ."
                        )
                    if value.shape[0] != self.data_arr.data.shape[0]:
                        raise ValueError(
                            "Length of weights not compatible with specified axis."
                        )
                self._weights = value
            else:
                raise TypeError("weights must be a numpy.ndarray.")
        else:
            self._weights = None

    @property
    def scaling(self):
        """
        Scaling factor used in combining images.

        Parameters
        ----------
        scale : function or `numpy.ndarray`-like or None, optional
            Images are multiplied by scaling prior to combining
            them. Scaling may be either a function, which will be applied to
            each image to determine the scaling factor, or a list or array
            whose length is the number of images in the `~ccdproc.Combiner`.
        """
        return self._scaling

    @scaling.setter
    def scaling(self, value):
        if value is None:
            self._scaling = value
        else:
            n_images = self.data_arr.data.shape[0]
            if callable(value):
                self._scaling = [value(self.data_arr[i]) for i in range(n_images)]
                self._scaling = np.array(self._scaling)
            else:
                try:
                    len(value)
                except TypeError as err:
                    raise TypeError(
                        "scaling must be a function or an array "
                        "the same length as the number of images.",
                    ) from err
                if len(value) != n_images:
                    raise ValueError(
                        "scaling must be a function or an array "
                        "the same length as the number of images."
                    )
                self._scaling = np.array(value)
            # reshape so that broadcasting occurs properly
            for _ in range(len(self.data_arr.data.shape) - 1):
                self._scaling = self.scaling[:, np.newaxis]

    # set up IRAF-like minmax clipping
    def clip_extrema(self, nlow=0, nhigh=0):
        """Mask pixels using an IRAF-like minmax clipping algorithm.  The
        algorithm will mask the lowest nlow values and the highest nhigh values
        before combining the values to make up a single pixel in the resulting
        image.  For example, the image will be a combination of
        Nimages-nlow-nhigh pixel values instead of the combination of Nimages.

        Parameters
        ----------
        nlow : int or None, optional
            If not None, the number of low values to reject from the
            combination.
            Default is 0.

        nhigh : int or None, optional
            If not None, the number of high values to reject from the
            combination.
            Default is 0.

        Notes
        -----
        Note that this differs slightly from the nominal IRAF imcombine
        behavior when other masks are in use.  For example, if ``nhigh>=1`` and
        any pixel is already masked for some other reason, then this algorithm
        will count the masking of that pixel toward the count of nhigh masked
        pixels.

        Here is a copy of the relevant IRAF help text [0]_:

        nlow = 1, nhigh = (minmax)
            The number of low and high pixels to be rejected by the "minmax"
            algorithm. These numbers are converted to fractions of the total
            number of input images so that if no rejections have taken place
            the specified number of pixels are rejected while if pixels have
            been rejected by masking, thresholding, or nonoverlap, then the
            fraction of the remaining pixels, truncated to an integer, is used.

        References
        ----------
        .. [0] image.imcombine help text.
           http://stsdas.stsci.edu/cgi-bin/gethelp.cgi?imcombine
        """

        if nlow is None:
            nlow = 0
        if nhigh is None:
            nhigh = 0

        argsorted = np.argsort(self.data_arr.data, axis=0)
        mg = np.mgrid[
            [slice(ndim) for i, ndim in enumerate(self.data_arr.shape) if i > 0]
        ]
        for i in range(-1 * nhigh, nlow):
            # create a tuple with the indices
            where = tuple([argsorted[i, :, :].ravel()] + [i.ravel() for i in mg])
            self.data_arr.mask[where] = True

    # set up min/max clipping algorithms
    def minmax_clipping(self, min_clip=None, max_clip=None):
        """Mask all pixels that are below min_clip or above max_clip.

        Parameters
        ----------
        min_clip : float or None, optional
            If not None, all pixels with values below min_clip will be masked.
            Default is ``None``.

        max_clip : float or None, optional
            If not None, all pixels with values above min_clip will be masked.
            Default is ``None``.
        """
        if min_clip is not None:
            mask = self.data_arr < min_clip
            self.data_arr.mask[mask] = True
        if max_clip is not None:
            mask = self.data_arr > max_clip
            self.data_arr.mask[mask] = True

    # set up sigma  clipping algorithms
    @deprecated_renamed_argument(
        "use_astropy",
        None,
        arg_in_kwargs=True,
        since="2.4.0",
        message="The use_astropy argument has been removed because "
        "astropy sigma clipping is now always used.",
    )
    def sigma_clipping(
        self, low_thresh=3, high_thresh=3, func="mean", dev_func="std", **kwd
    ):
        """
        Pixels will be rejected if they have deviations greater than those
        set by the threshold values. The algorithm will first calculated
        a baseline value using the function specified in func and deviation
        based on dev_func and the input data array. Any pixel with a
        deviation from the baseline value greater than that set by
        high_thresh or lower than that set by low_thresh will be rejected.

        Parameters
        ----------
        low_thresh : positive float or None, optional
            Threshold for rejecting pixels that deviate below the baseline
            value. If negative value, then will be convert to a positive
            value. If None, no rejection will be done based on low_thresh.
            Default is 3.

        high_thresh : positive float or None, optional
            Threshold for rejecting pixels that deviate above the baseline
            value. If None, no rejection will be done based on high_thresh.
            Default is 3.

        func : {'median', 'mean'} or callable, optional
            The statistic or callable function/object used to compute
            the center value for the clipping. If using a callable
            function/object and the ``axis`` keyword is used, then it must
            be able to ignore NaNs (e.g., `numpy.nanmean`) and it must have
            an ``axis`` keyword to return an array with axis dimension(s)
            removed. The default is ``'median'``.

        dev_func : {'std', 'mad_std'} or callable, optional
            The statistic or callable function/object used to compute the
            standard deviation about the center value. If using a callable
            function/object and the ``axis`` keyword is used, then it must
            be able to ignore NaNs (e.g., `numpy.nanstd`) and it must have
            an ``axis`` keyword to return an array with axis dimension(s)
            removed. The default is ``'std'``.

        kwd
            Any remaining keyword arguments are passed to astropy's
            :func:`~astropy.stats.sigma_clip` function.
        """

        # Remove in 3.0
        _ = kwd.pop("use_astropy", True)

        self.data_arr.mask |= sigma_clip(
            self.data_arr.data,
            sigma_lower=low_thresh,
            sigma_upper=high_thresh,
            axis=kwd.get("axis", 0),
            copy=kwd.get("copy", False),
            maxiters=kwd.get("maxiters", 1),
            cenfunc=func,
            stdfunc=dev_func,
            masked=True,
            **kwd,
        ).mask

    def _get_scaled_data(self, scale_arg):
        if scale_arg is not None:
            return self.data_arr * scale_arg
        if self.scaling is not None:
            return self.data_arr * self.scaling
        return self.data_arr

    def _get_nan_substituted_data(self, data):
        # Get the data as an unmasked array with masked values filled as NaN
        if self.data_arr.mask.any():
            data = np.ma.filled(data, fill_value=np.nan)
        else:
            data = data.data
        return data

    def _combination_setup(self, user_func, default_func, scale_to):
        """
        Handle the common pieces of image combination data/mask setup.
        """
        data = self._get_scaled_data(scale_to)

        # Play it safe for now and only do the nan thing if the user is using
        # the default combination function.
        if user_func is None:
            combo_func = default_func
            # Subtitute NaN for masked entries
            data = self._get_nan_substituted_data(data)
            masked_values = np.isnan(data).sum(axis=0)
        else:
            masked_values = self.data_arr.mask.sum(axis=0)
            combo_func = user_func

        return data, masked_values, combo_func

    # set up the combining algorithms
    def median_combine(
        self, median_func=None, scale_to=None, uncertainty_func=sigma_func
    ):
        """
        Median combine a set of arrays.

        A `~astropy.nddata.CCDData` object is returned with the data property set to
        the median of the arrays. If the data was masked or any data have been
        rejected, those pixels will not be included in the median. A mask will
        be returned, and if a pixel has been rejected in all images, it will be
        masked. The uncertainty of the combined image is set by 1.4826 times
        the median absolute deviation of all input images.

        Parameters
        ----------
        median_func : function, optional
            Function that calculates median of a `numpy.ma.MaskedArray`.
            Default is `numpy.ma.median`.

        scale_to : float or None, optional
            Scaling factor used in the average combined image. If given,
            it overrides `scaling`.
            Defaults to None.

        uncertainty_func : function, optional
            Function to calculate uncertainty.
            Defaults is `~ccdproc.sigma_func`.

        Returns
        -------
        combined_image: `~astropy.nddata.CCDData`
            CCDData object based on the combined input of CCDData objects.

        Warnings
        --------
        The uncertainty currently calculated using the median absolute
        deviation does not account for rejected pixels.
        """

        data, masked_values, median_func = self._combination_setup(
            median_func, _default_median(), scale_to
        )

        medianed = median_func(data, axis=0)

        # set the mask
        mask = masked_values == len(self.data_arr)

        # set the uncertainty

        # This still uses numpy for the median because the astropy
        # code requires that the median function take the argument
        # overwrite_input and bottleneck doesn't allow that argument.
        # This is ugly, but setting ignore_nan to True should make sure
        # that either nans or masks are handled properly.
        if uncertainty_func is sigma_func:
            uncertainty = uncertainty_func(data, axis=0, ignore_nan=True)
        else:
            uncertainty = uncertainty_func(data, axis=0)
        # Divide uncertainty by the number of pixel (#309)
        uncertainty /= np.sqrt(len(self.data_arr) - masked_values)
        # Convert uncertainty to plain numpy array (#351)
        # There is no need to care about potential masks because the
        # uncertainty was calculated based on the data so potential masked
        # elements are also masked in the data. No need to keep two identical
        # masks.
        uncertainty = np.asarray(uncertainty)

        # create the combined image with a dtype matching the combiner
        combined_image = CCDData(
            np.asarray(medianed, dtype=self.dtype),
            mask=mask,
            unit=self.unit,
            uncertainty=StdDevUncertainty(uncertainty),
        )

        # update the meta data
        combined_image.meta["NCOMBINE"] = len(self.data_arr)

        # return the combined image
        return combined_image

    def _weighted_sum(self, data, sum_func):
        """
        Perform weighted sum, used by both ``sum_combine`` and in some cases
        by ``average_combine``.
        """
        if self.weights.shape != data.shape:
            # Add extra axes to the weights for broadcasting
            weights = np.reshape(self.weights, [len(self.weights), 1, 1])
        else:
            weights = self.weights

        # Turns out bn.nansum has an implementation that is not
        # precise enough for float32 sums. Doing this should
        # ensure the sums are carried out as float64
        weights = weights.astype("float64")
        weighted_sum = sum_func(data * weights, axis=0)
        return weighted_sum, weights

    def average_combine(
        self,
        scale_func=None,
        scale_to=None,
        uncertainty_func=_default_std_dev_func,
        sum_func=_default_sum_func,
    ):
        """
        Average combine together a set of arrays.

        A `~astropy.nddata.CCDData` object is returned with the data property
        set to the average of the arrays. If the data was masked or any
        data have been rejected, those pixels will not be included in the
        average. A mask will be returned, and if a pixel has been
        rejected in all images, it will be masked. The uncertainty of
        the combined image is set by the standard deviation of the input
        images.

        Parameters
        ----------
        scale_func : function, optional
            Function to calculate the average. Defaults to
            `numpy.nanmean`.

        scale_to : float or None, optional
            Scaling factor used in the average combined image. If given,
            it overrides `scaling`. Defaults to ``None``.

        uncertainty_func : function, optional
            Function to calculate uncertainty. Defaults to `numpy.ma.std`.

        sum_func : function, optional
            Function used to calculate sums, including the one done to
            find the weighted average. Defaults to `numpy.nansum`.

        Returns
        -------
        combined_image: `~astropy.nddata.CCDData`
            CCDData object based on the combined input of CCDData objects.
        """
        data, masked_values, scale_func = self._combination_setup(
            scale_func, _default_average(), scale_to
        )

        # Do NOT modify data after this -- we need it to be intact when we
        # we get to the uncertainty calculation.
        if self.weights is not None:
            weighted_sum, weights = self._weighted_sum(data, sum_func)
            mean = weighted_sum / sum_func(weights, axis=0)
        else:
            mean = scale_func(data, axis=0)

        # calculate the mask

        mask = masked_values == len(self.data_arr)

        # set up the deviation
        uncertainty = uncertainty_func(data, axis=0)
        # Divide uncertainty by the number of pixel (#309)
        uncertainty /= np.sqrt(len(data) - masked_values)
        # Convert uncertainty to plain numpy array (#351)
        uncertainty = np.asarray(uncertainty)

        # create the combined image with a dtype that matches the combiner
        combined_image = CCDData(
            np.asarray(mean, dtype=self.dtype),
            mask=mask,
            unit=self.unit,
            uncertainty=StdDevUncertainty(uncertainty),
        )

        # update the meta data
        combined_image.meta["NCOMBINE"] = len(data)

        # return the combined image
        return combined_image

    def sum_combine(
        self, sum_func=None, scale_to=None, uncertainty_func=_default_std_dev_func
    ):
        """
        Sum combine together a set of arrays.

        A `~astropy.nddata.CCDData` object is returned with the data property
        set to the sum of the arrays. If the data was masked or any
        data have been rejected, those pixels will not be included in the
        sum. A mask will be returned, and if a pixel has been
        rejected in all images, it will be masked. The uncertainty of
        the combined image is set by the multiplication of summation of
        standard deviation of the input by square root of number of images.
        Because sum_combine returns 'pure sum' with masked pixels ignored, if
        re-scaled sum is needed, average_combine have to be used with
        multiplication by number of images combined.

        Parameters
        ----------
        sum_func : function, optional
            Function to calculate the sum. Defaults to
            `numpy.nansum` or ``bottleneck.nansum``.

        scale_to : float or None, optional
            Scaling factor used in the sum combined image. If given,
            it overrides `scaling`. Defaults to ``None``.

        uncertainty_func : function, optional
            Function to calculate uncertainty. Defaults to `numpy.ma.std`.

        Returns
        -------
        combined_image: `~astropy.nddata.CCDData`
            CCDData object based on the combined input of CCDData objects.
        """

        data, masked_values, sum_func = self._combination_setup(
            sum_func, _default_sum(), scale_to
        )

        if self.weights is not None:
            summed, weights = self._weighted_sum(data, sum_func)
        else:
            summed = sum_func(data, axis=0)

        # set up the mask
        mask = masked_values == len(self.data_arr)

        # set up the deviation
        uncertainty = uncertainty_func(data, axis=0)
        # Divide uncertainty by the number of pixel (#309)
        uncertainty /= np.sqrt(len(data) - masked_values)
        # Convert uncertainty to plain numpy array (#351)
        uncertainty = np.asarray(uncertainty)
        # Multiply uncertainty by square root of the number of images
        uncertainty *= len(data) - masked_values

        # create the combined image with a dtype that matches the combiner
        combined_image = CCDData(
            np.asarray(summed, dtype=self.dtype),
            mask=mask,
            unit=self.unit,
            uncertainty=StdDevUncertainty(uncertainty),
        )

        # update the meta data
        combined_image.meta["NCOMBINE"] = len(self.data_arr)

        # return the combined image
        return combined_image


def _calculate_step_sizes(x_size, y_size, num_chunks):
    """
    Calculate the strides in x and y to achieve at least
    the ``num_chunks`` pieces.

    Parameters
    ----------
    """
    # First we try to split only along fast x axis
    xstep = max(1, int(x_size / num_chunks))

    # More chunks are needed only if xstep gives us fewer chunks than
    # requested.
    x_chunks = int(x_size / xstep)

    if x_chunks >= num_chunks:
        ystep = y_size
    else:
        # The x and y loops are nested, so the number of chunks
        # is multiplicative, not additive. Calculate the number
        # of y chunks we need to get at num_chunks.
        y_chunks = int(num_chunks / x_chunks) + 1
        ystep = max(1, int(y_size / y_chunks))

    return xstep, ystep


def _calculate_size_of_image(ccd, combine_uncertainty_function):
    # If uncertainty_func is given for combine this will create an uncertainty
    # even if the originals did not have one. In that case we need to create
    # an empty placeholder.
    if ccd.uncertainty is None and combine_uncertainty_function is not None:
        ccd.uncertainty = StdDevUncertainty(np.zeros(ccd.data.shape))

    size_of_an_img = ccd.data.nbytes
    try:
        size_of_an_img += ccd.uncertainty.array.nbytes
    # In case uncertainty is None it has no "array" and in case the "array" is
    # not a numpy array:
    except AttributeError:
        pass
    # Mask is enforced to be a numpy.array across astropy versions
    if ccd.mask is not None:
        size_of_an_img += ccd.mask.nbytes
    # flags is not necessarily a numpy array so do not fail with an
    # AttributeError in case something was set!
    # TODO: Flags are not taken into account in Combiner. This number is added
    #       nevertheless for future compatibility.
    try:
        size_of_an_img += ccd.flags.nbytes
    except AttributeError:
        pass

    return size_of_an_img


def combine(
    img_list,
    output_file=None,
    method="average",
    weights=None,
    scale=None,
    mem_limit=16e9,
    clip_extrema=False,
    nlow=1,
    nhigh=1,
    minmax_clip=False,
    minmax_clip_min=None,
    minmax_clip_max=None,
    sigma_clip=False,
    sigma_clip_low_thresh=3,
    sigma_clip_high_thresh=3,
    sigma_clip_func=ma.mean,
    sigma_clip_dev_func=ma.std,
    dtype=None,
    combine_uncertainty_function=None,
    overwrite_output=False,
    **ccdkwargs,
):
    """
    Convenience function for combining multiple images.

    Parameters
    ----------
    img_list : `numpy.ndarray`, list or str
        A list of fits filenames or `~astropy.nddata.CCDData` objects that will be
        combined together. Or a string of fits filenames separated by comma
        ",".

    output_file : str or None, optional
        Optional output fits file-name to which the final output can be
        directly written.
        Default is ``None``.

    method : str, optional
        Method to combine images:

        - ``'average'`` : To combine by calculating the average.
        - ``'median'`` : To combine by calculating the median.
        - ``'sum'`` : To combine by calculating the sum.

        Default is ``'average'``.

    weights : `numpy.ndarray` or None, optional
        Weights to be used when combining images.
        An array with the weight values. The dimensions should match the
        the dimensions of the data arrays being combined.
        Default is ``None``.

    scale : function or `numpy.ndarray`-like or None, optional
        Scaling factor to be used when combining images.
        Images are multiplied by scaling prior to combining them. Scaling
        may be either a function, which will be applied to each image
        to determine the scaling factor, or a list or array whose length
        is the number of images in the `Combiner`. Default is ``None``.

    mem_limit : float, optional
        Maximum memory which should be used while combining (in bytes).
        Default is ``16e9``.

    clip_extrema : bool, optional
        Set to True if you want to mask pixels using an IRAF-like minmax
        clipping algorithm.  The algorithm will mask the lowest nlow values and
        the highest nhigh values before combining the values to make up a
        single pixel in the resulting image.  For example, the image will be a
        combination of Nimages-low-nhigh pixel values instead of the
        combination of Nimages.

        Parameters below are valid only when clip_extrema is set to True,
        see :meth:`Combiner.clip_extrema` for the parameter description:

        - ``nlow`` : int or None, optional
        - ``nhigh`` : int or None, optional

    minmax_clip : bool, optional
        Set to True if you want to mask all pixels that are below
        minmax_clip_min or above minmax_clip_max before combining.
        Default is ``False``.

        Parameters below are valid only when minmax_clip is set to True, see
        :meth:`Combiner.minmax_clipping` for the parameter description:

        - ``minmax_clip_min`` : float or None, optional
        - ``minmax_clip_max`` : float or None, optional

    sigma_clip : bool, optional
        Set to True if you want to reject pixels which have deviations greater
        than those
        set by the threshold values. The algorithm will first calculated
        a baseline value using the function specified in func and deviation
        based on sigma_clip_dev_func and the input data array. Any pixel with
        a deviation from the baseline value greater than that set by
        sigma_clip_high_thresh or lower than that set by sigma_clip_low_thresh
        will be rejected.
        Default is ``False``.

        Parameters below are valid only when sigma_clip is set to True. See
        :meth:`Combiner.sigma_clipping` for the parameter description.

        - ``sigma_clip_low_thresh`` : positive float or None, optional
        - ``sigma_clip_high_thresh`` : positive float or None, optional
        - ``sigma_clip_func`` : function, optional
        - ``sigma_clip_dev_func`` : function, optional

    dtype : str or `numpy.dtype` or None, optional
        The intermediate and resulting ``dtype`` for the combined CCDs. See
        `ccdproc.Combiner`. If ``None`` this is set to ``float64``.
        Default is ``None``.

    combine_uncertainty_function : callable, None, optional
        If ``None`` use the default uncertainty func when using average, median or
        sum combine, otherwise use the function provided.
        Default is ``None``.

    overwrite_output : bool, optional
        If ``output_file`` is specified, this is passed to the
        `astropy.nddata.fits_ccddata_writer` under the keyword ``overwrite``;
        has no effect otherwise.
        Default is ``False``.

    ccdkwargs : Other keyword arguments for `astropy.nddata.fits_ccddata_reader`.

    Returns
    -------
    combined_image : `~astropy.nddata.CCDData`
        CCDData object based on the combined input of CCDData objects.
    """
    if not isinstance(img_list, list):
        # If not a list, check whether it is a numpy ndarray or string of
        # filenames separated by comma
        if isinstance(img_list, np.ndarray):
            img_list = img_list.tolist()
        elif isinstance(img_list, str) and ("," in img_list):
            img_list = img_list.split(",")
        else:
            try:
                # Maybe the input can be made into a list, so try that
                img_list = list(img_list)
            except TypeError as err:
                raise ValueError(
                    "unrecognised input for list of images to combine."
                ) from err

    # Select Combine function to call in Combiner
    if method == "average":
        combine_function = "average_combine"
    elif method == "median":
        combine_function = "median_combine"
    elif method == "sum":
        combine_function = "sum_combine"
    else:
        raise ValueError(f"unrecognised combine method : {method}.")

    # First we create a CCDObject from first image for storing output
    if isinstance(img_list[0], CCDData):
        ccd = img_list[0].copy()
    else:
        # User has provided fits filenames to read from
        ccd = CCDData.read(img_list[0], **ccdkwargs)

    if dtype is None:
        dtype = np.float64

    # Convert the master image to the appropriate dtype so when overwriting it
    # later the data is not downcast and the memory consumption calculation
    # uses the internally used dtype instead of the original dtype. #391
    if ccd.data.dtype != dtype:
        ccd.data = ccd.data.astype(dtype)

    # If the template image doesn't have an uncertainty, add one, because the
    # result always has an uncertainty.
    if ccd.uncertainty is None:
        ccd.uncertainty = StdDevUncertainty(np.zeros_like(ccd.data))

    # If the template doesn't have a mask, add one, because the result may have
    # a mask
    if ccd.mask is None:
        ccd.mask = np.zeros_like(ccd.data, dtype=bool)

    size_of_an_img = _calculate_size_of_image(ccd, combine_uncertainty_function)

    no_of_img = len(img_list)

    # Set a memory use factor based on profiling
    if method == "median":
        memory_factor = 3
    else:
        memory_factor = 2

    memory_factor *= 1.3

    # determine the number of chunks to split the images into
    no_chunks = int((memory_factor * size_of_an_img * no_of_img) / mem_limit) + 1
    if no_chunks > 1:
        log.info(
            f"splitting each image into {no_chunks} chunks to limit memory usage "
            f"to {mem_limit} bytes."
        )
    xs, ys = ccd.data.shape

    # Calculate strides for loop
    xstep, ystep = _calculate_step_sizes(xs, ys, no_chunks)

    # Dictionary of Combiner properties to set and methods to call before
    # combining
    to_set_in_combiner = {}
    to_call_in_combiner = {}

    # Define all the Combiner properties one wants to apply before combining
    # images
    if weights is not None:
        to_set_in_combiner["weights"] = weights

    if scale is not None:
        # If the scale is a function, then scaling function need to be applied
        # on full image to obtain scaling factor and create an array instead.
        if callable(scale):
            scalevalues = []
            for image in img_list:
                if isinstance(image, CCDData):
                    imgccd = image
                else:
                    imgccd = CCDData.read(image, **ccdkwargs)

                scalevalues.append(scale(imgccd.data))

            to_set_in_combiner["scaling"] = np.array(scalevalues)
        else:
            to_set_in_combiner["scaling"] = scale

    if clip_extrema:
        to_call_in_combiner["clip_extrema"] = {"nlow": nlow, "nhigh": nhigh}

    if minmax_clip:
        to_call_in_combiner["minmax_clipping"] = {
            "min_clip": minmax_clip_min,
            "max_clip": minmax_clip_max,
        }

    if sigma_clip:
        to_call_in_combiner["sigma_clipping"] = {
            "low_thresh": sigma_clip_low_thresh,
            "high_thresh": sigma_clip_high_thresh,
            "func": sigma_clip_func,
            "dev_func": sigma_clip_dev_func,
        }

    # Finally Run the input method on all the subsections of the image
    # and write final stitched image to ccd
    for x in range(0, xs, xstep):
        for y in range(0, ys, ystep):
            xend, yend = min(xs, x + xstep), min(ys, y + ystep)
            ccd_list = []
            for image in img_list:
                if isinstance(image, CCDData):
                    imgccd = image
                else:
                    imgccd = CCDData.read(image, **ccdkwargs)

                # Trim image and copy
                # The copy is *essential* to avoid having a bunch
                # of unused file references around if the files
                # are memory-mapped. See this PR for details
                # https://github.com/astropy/ccdproc/pull/630
                ccd_list.append(imgccd[x:xend, y:yend].copy())

            # Create Combiner for tile
            tile_combiner = Combiner(ccd_list, dtype=dtype)

            # Set all properties and call all methods
            for to_set in to_set_in_combiner:
                setattr(tile_combiner, to_set, to_set_in_combiner[to_set])
            for to_call in to_call_in_combiner:
                getattr(tile_combiner, to_call)(**to_call_in_combiner[to_call])

            # Finally call the combine algorithm
            combine_kwds = {}
            if combine_uncertainty_function is not None:
                combine_kwds["uncertainty_func"] = combine_uncertainty_function

            comb_tile = getattr(tile_combiner, combine_function)(**combine_kwds)

            # add it back into the master image
            ccd.data[x:xend, y:yend] = comb_tile.data
            if ccd.mask is not None:
                ccd.mask[x:xend, y:yend] = comb_tile.mask
            if ccd.uncertainty is not None:
                ccd.uncertainty.array[x:xend, y:yend] = comb_tile.uncertainty.array
            # Free up memory to try to stay under user's limit
            del comb_tile
            del tile_combiner
            del ccd_list

    # Write fits file if filename was provided
    if output_file is not None:
        ccd.write(output_file, overwrite=overwrite_output)

    return ccd



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This module implements the base CCDData class."""

from astropy.nddata import CCDData, fits_ccddata_reader, fits_ccddata_writer

__all__ = ["CCDData", "fits_ccddata_reader", "fits_ccddata_writer"]


# This should be be a tuple to ensure it isn't inadvertently changed
# elsewhere.
_recognized_fits_file_extensions = ("fit", "fits", "fts")



import gc
import mmap
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
from astropy.io import fits

# This bit of hackery ensures that we can see ccdproc from within
# the test suite
sys.path.append(str(Path().cwd()))
from ccdproc import combine

# Do not combine these into one statement. When all references are lost
# to a TemporaryDirectory the directory is automatically deleted. _TMPDIR
# creates a reference that will stick around.
_TMPDIR = TemporaryDirectory()
TMPPATH = Path(_TMPDIR.name)

ALLOWED_EXTENSIONS = {"fits": "fits", "plain": "txt"}


def generate_fits_files(number, size=None):
    if size is None:
        use_size = [250, 250]
    else:
        int_size = int(size)
        use_size = [int_size, int_size]

    base_name = "test-combine-{num:03d}." + ALLOWED_EXTENSIONS["fits"]

    for num in range(number):
        data = np.zeros(shape=use_size)
        hdu = fits.PrimaryHDU(data=data)
        hdu.header["bunit"] = "adu"
        name = base_name.format(num=num)
        path = TMPPATH / name
        hdu.writeto(path, overwrite=True)


def generate_plain_files(number):
    for i in range(number):
        file = TMPPATH / (f"{i:03d}." + ALLOWED_EXTENSIONS["plain"])
        file.write_bytes(np.random.default_rng().random(100))


def open_files_with_open(kind):
    """
    Open files with plain open.
    """
    # Ensure the file references persist until end of script. Not really
    # necessary, but convenient while debugging the script.
    global fds
    fds = []

    paths = TMPPATH.glob("**/*." + ALLOWED_EXTENSIONS[kind])

    for p in paths:
        fds.append(p.open())


def open_files_as_mmap(kind):
    """
    Open files as mmaps.
    """
    # Ensure the file references persist until end of script. Not really
    # necessary, but convenient while debugging the script.
    global fds
    fds = []

    paths = TMPPATH.glob("**/*." + ALLOWED_EXTENSIONS[kind])

    for p in paths:
        with p.open() as f:
            fds.append(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_COPY))


def open_files_ccdproc_combine_chunk(kind):
    """
    Open files indirectly as part of ccdproc.combine, ensuring that the
    task is broken into chunks.
    """
    global combo
    paths = sorted(list(TMPPATH.glob("**/*." + ALLOWED_EXTENSIONS[kind])))
    # We want to force combine to break the task into chunks even
    # if the task really would fit in memory; it is in that case that
    # we end up with too many open files. We'll open one file, determine
    # the size of the data in bytes, and set the memory limit to that.
    # That will mean lots of chunks (however many files there are plus one),
    # but lots of chunks is fine.
    with fits.open(paths[0]) as hdulist:
        array_size = hdulist[0].data.nbytes

    combo = combine(paths, mem_limit=array_size)


def open_files_ccdproc_combine_nochunk(kind):
    """
    Open files indirectly as part of ccdproc.combine, ensuring that the
    task is not broken into chunks.
    """
    global combo
    paths = sorted(list(TMPPATH.glob("**/*." + ALLOWED_EXTENSIONS[kind])))

    # We ensure there are no chunks by setting a memory limit large
    # enough to hold everything.
    with fits.open(paths[0]) as hdulist:
        array_size = hdulist[0].data.nbytes

    # Why 2x the number of files? To make absolutely sure we don't
    # end up chunking the job.
    array_size *= 2 * len(paths)
    combo = combine(paths)


ALLOWED_OPENERS = {
    "open": open_files_with_open,
    "mmap": open_files_as_mmap,
    "combine-chunk": open_files_ccdproc_combine_chunk,
    "combine-nochunk": open_files_ccdproc_combine_nochunk,
}


def run_with_limit(n, kind="fits", size=None, overhead=6, open_method="mmap"):
    """
    Try opening a bunch of files with a relatively low limit on the number
    of open files.

    Parameters
    ----------

    n : int
        Limit on number of open files in this function. The number of files
        to create is calculated from this to be just below the maximum number
        of files controlled by this function that can be opened.

    kind : one of 'fits', 'plain', optional
        The type of file to generate. The plain files are intended mainly for
        testing this script, while the FITS files are for testing
        ccdproc.combine.

    size : int, optional
        Size of file to create. If the kind is 'plain; this is the size
        of the file, in bytes. If the kind is 'fits', this is the size
        of one side of the image (the image is always square).

    overhead : int, optional
        Number of open files to assume the OS is using for this process. The
        default value is chosen so that this succeeds on MacOS or Linux.
        Setting it to a value lower than default should cause a SystemExit
        exception to be raised because of too many open files. This is meant
        for testing that this script is actually testing something.

    Notes
    -----

    .. warning::

        You should run this in a subprocess. Running as part of a larger python
        process will lower the limit on the number of open files for that
        **entire python process** which will almost certainly lead to nasty
        side effects.
    """
    # Keep the resource import here so that it is skipped on windows
    import resource

    # Do a little input validation
    if n <= 0:
        raise ValueError("Argument 'n' must be a positive integer")

    if kind not in ALLOWED_EXTENSIONS.keys():
        raise ValueError(
            "Argument 'kind' must be one of " f"{ALLOWED_EXTENSIONS.keys()}"
        )

    # Set the limit on the number of open files to n. The try/except
    # is the catch the case where this change would *increase*, rather than
    # decrease, the limit. That apparently can only be done by a superuser.
    try:
        resource.setrlimit(resource.RLIMIT_NOFILE, (n, n))
    except ValueError as e:
        if "not allowed to raise maximum limit" not in str(e):
            raise
        max_n_this_process = resource.getrlimit(resource.RLIMIT_NOFILE)
        raise ValueError(
            "Maximum number of open " f"files is {max_n_this_process}"
        ) from e

    # The "-1" is to leave a little wiggle room. overhead is based on the
    # the number of open files that a process running on linux has open.
    # These typically include stdin and stout, and apparently others.
    n_files = n - 1 - overhead

    proc = psutil.Process()

    print("Process ID is: ", proc.pid, flush=True)
    print(f"Making {n_files} files")
    if kind == "plain":
        generate_plain_files(n_files)
    elif kind == "fits":
        generate_fits_files(n_files, size=size)

    # Print number of open files before we try opening anything for debugging
    # purposes.
    print(f"Before opening, files open is {len(proc.open_files())}", flush=True)
    print("    Note well: this number is different than what lsof reports.")

    try:
        ALLOWED_OPENERS[open_method](kind)
        # fds.append(p.open())
    except OSError as e:
        # Capture the error and re-raise as a SystemExit because this is
        # run in a subprocess. This ensures that the original error message
        # is reported back to the calling process; we add on the number of
        # open files.
        raise SystemExit(
            str(e)
            + "; number of open files: "
            + f"{len(proc.open_files())}, with target {n_files}"
        ) from e
    else:
        print(
            "Opens succeeded, files currently open:", len(proc.open_files()), flush=True
        )


if __name__ == "__main__":
    from argparse import ArgumentParser

    import psutil

    parser = ArgumentParser()
    parser.add_argument("number", type=int, help="Limit on number of open files.")
    parser.add_argument(
        "--kind",
        action="store",
        default="plain",
        choices=ALLOWED_EXTENSIONS.keys(),
        help="Kind of file to generate for test; " "default is plain",
    )
    parser.add_argument(
        "--overhead",
        type=int,
        action="store",
        help="Number of files to assume the OS is using.",
        default=6,
    )
    parser.add_argument(
        "--open-by",
        action="store",
        default="mmap",
        choices=ALLOWED_OPENERS.keys(),
        help="How to open the files. Default is mmap",
    )
    parser.add_argument(
        "--size",
        type=int,
        action="store",
        help="Size of one side of image to create. "
        "All images are square, so only give "
        "a single number for the size.",
    )
    parser.add_argument(
        "--frequent-gc",
        action="store_true",
        help="If set, perform garbage collection "
        "much more frequently than the default.",
    )
    args = parser.parse_args()
    if args.frequent_gc:
        gc.set_threshold(10, 10, 10)
    print("Garbage collection thresholds: ", gc.get_threshold())
    run_with_limit(
        args.number,
        kind=args.kind,
        overhead=args.overhead,
        open_method=args.open_by,
        size=args.size,
    )
